# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
from typing import Any, Optional

import tqdm


class ProgressMonitor(abc.ABC):
    """Progress Monitor for file upload"""

    @abc.abstractmethod
    def update(self, delta: int):
        raise NotImplementedError("update")

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError("close")

    @abc.abstractmethod
    def is_closed(self) -> bool:
        raise NotImplementedError("is_closed")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class ProgressMonitorFactory(abc.ABC):
    """Progress Monitor Factory for file upload"""

    __ctx: dict[str, Any]

    def __init__(self, ctx: Optional[dict[str, Any]] = None):
        self.__ctx = ctx or {}

    @abc.abstractmethod
    def upload_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        raise NotImplementedError("upload_monitor")

    @abc.abstractmethod
    def download_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        raise NotImplementedError("download_monitor")

    def get_context(self) -> dict[str, Any]:
        return self.__ctx


class NoopProgressMonitor(ProgressMonitor):
    """A Noop Progress Monitor"""

    def update(self, uploaded_bytes: int):
        pass

    def close(self):
        pass

    def is_closed(self) -> bool:
        return False


class NoopProgressMonitorFactory(ProgressMonitorFactory):
    """A Noop Progress Monitor Factory"""

    def upload_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        return NoopProgressMonitor()

    def download_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        return NoopProgressMonitor()


class TqdmProgressMonitor(ProgressMonitor):
    """A Tqdm Progress Monitor"""

    __tqdm: tqdm.tqdm
    __is_closed: bool
    __total: int
    __auto_close: bool
    __uploaded_bytes: int

    def __init__(
        self,
        total: int,
        desc: str,
        position: int = 0,
        leave: bool = True,
        unit: Optional[str] = "B",
    ):
        tqdm_args = {
            "total": total,
            "desc": desc,
            "bar_format": "{percentage:.1f}%|{bar:25} | {n_fmt}/{total_fmt} | {rate_fmt} | {elapsed} | {desc}",
            "position": position,
            "leave": leave,
        }

        if unit is not None:
            tqdm_args["unit"] = unit
            if unit == "B":
                tqdm_args["unit_scale"] = True
                tqdm_args["unit_divisor"] = 1024

        self.__tqdm = tqdm.tqdm(**tqdm_args)
        self.__is_closed = False

    def update(self, uploaded_bytes: int):
        self.__tqdm.update(uploaded_bytes)

    def close(self):
        self.__tqdm.close()
        self.__is_closed = True

    def is_closed(self) -> bool:
        return self.__is_closed


class TqdmProgressMonitorFactory(ProgressMonitorFactory):
    """A Tqdm Progress Monitor Factory"""

    __monitors: list[Optional[ProgressMonitor]]

    def __init__(
        self,
        concurrency: int = 1,
        ctx: Optional[dict[str, Any]] = None,
        unit: Optional[str] = "B",
    ):
        super().__init__(ctx=ctx)
        self.__monitors = [None] * concurrency

    def __first_available_slot(self) -> Optional[int]:
        for idx in range(len(self.__monitors)):
            if self.__monitors[idx] is None or self.__monitors[idx].is_closed():  # type: ignore[union-attr]
                return idx

        return None

    def __any_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        # This for sure is not fully threadsafe, but it 100% works for single threading and
        # _mostly_ works for multithreading.
        slot = self.__first_available_slot()
        if slot is None:
            raise ValueError("Number of concurrent monitors is exceeding concurrency!")

        progress_monitor_args = {
            "total": size,
            "desc": f"Src: {source}",
            "position": slot,
            "leave": len(self.__monitors) == 1,
        }

        if kwargs:
            progress_monitor_args.update(**kwargs)

        monitor = TqdmProgressMonitor(**progress_monitor_args)  # type: ignore[arg-type]

        self.__monitors[slot] = monitor

        return monitor

    def upload_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        return self.__any_monitor(source=source, size=size, kwargs=kwargs)

    def download_monitor(
        self, source: str, size: int, kwargs: Optional[dict[str, Any]] = None
    ) -> ProgressMonitor:
        return self.__any_monitor(source=source, size=size, kwargs=kwargs)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Any, Optional, Union

import pydantic
from pydantic import ConfigDict

from ...sentinels import (
    NotSet,
    NotSetType,
    is_not_set,
)
from ...updates import MetadataChangeset
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionReference,
    ComputeRequirements,
    ContainerParameters,
)


class CreateActionRequest(pydantic.BaseModel):
    """
    Request payload to create a new action
    """

    # Required
    name: str

    # Optional
    compute_requirements: Optional[ComputeRequirements] = None
    container_parameters: Optional[ContainerParameters] = None
    description: Optional[str] = None
    inherits: Optional[ActionReference] = None
    metadata: Optional[dict[str, Any]] = pydantic.Field(default_factory=dict)
    parameters: Optional[list[ActionParameter]] = pydantic.Field(default_factory=list)
    short_description: Optional[str] = None
    tags: Optional[list[str]] = pydantic.Field(default_factory=list)
    timeout: Optional[int] = None
    uri: Optional[str] = None


class SetActionAccessibilityRequest(pydantic.BaseModel):
    """
    Request payload to set action accessibility
    """

    accessibility: Accessibility
    digest: Optional[str] = None
    """Specify specific version of Action. If not specified, the latest version's accessibility will be updated."""
    model_config = ConfigDict(extra="forbid")


class UpdateActionRequest(pydantic.BaseModel):
    """
    Request payload to update an action
    """

    compute_requirements: Optional[Union[ComputeRequirements, NotSetType]] = NotSet
    container_parameters: Optional[Union[ContainerParameters, NotSetType]] = NotSet
    description: Optional[Union[str, NotSetType]] = NotSet
    inherits: Optional[Union[ActionReference, NotSetType]] = NotSet
    metadata_changeset: Union[MetadataChangeset, NotSetType] = NotSet
    parameter_changeset: Union[ActionParameterChangeset, NotSetType] = NotSet
    uri: Optional[Union[str, NotSetType]] = NotSet
    short_description: Optional[Union[str, NotSetType]] = NotSet
    timeout: Optional[Union[int, NotSetType]] = NotSet

    @pydantic.field_validator("uri")
    def validate_uri(cls, v):
        if v is None or is_not_set(v):
            return v

        stripped = v.strip()
        if not stripped:
            raise ValueError("uri cannot be empty")
        return stripped

    model_config = ConfigDict(
        extra="forbid", json_schema_extra=NotSetType.openapi_schema_modifier
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .http_resources import (
    ContainerUploadCredentials,
    CreateImageRepositoryRequest,
    CreateImageRepositoryResponse,
    DeleteImageRepositoryRequest,
    DeleteImageRequest,
    RepositoryContainsImageResponse,
    SetImageRepositoryImmutableTagsRequest,
)
from .image_registry import (
    ContainerCredentials,
    ImageRegistry,
    ImageRepository,
    RepositoryPurpose,
    RepositoryTag,
)
from .record import (
    ContainerImageRecord,
    ContainerImageRepositoryRecord,
)

__all__ = (
    "ContainerCredentials",
    "ContainerImageRecord",
    "ContainerImageRepositoryRecord",
    "ContainerUploadCredentials",
    "CreateImageRepositoryRequest",
    "CreateImageRepositoryResponse",
    "DeleteImageRequest",
    "DeleteImageRepositoryRequest",
    "ImageRegistry",
    "ImageRepository",
    "RepositoryContainsImageResponse",
    "RepositoryPurpose",
    "RepositoryTag",
    "SetImageRepositoryImmutableTagsRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import typing
import urllib.parse

import pydantic

from ..auth import Permissions
from ..http import PaginatedList, RobotoClient
from ..time import utcnow
from .http_resources import (
    CreateImageRepositoryRequest,
    DeleteImageRepositoryRequest,
    DeleteImageRequest,
)
from .record import (
    ContainerImageRecord,
    ContainerImageRepositoryRecord,
)


class ContainerCredentials(pydantic.BaseModel):
    """Credentials used by a container"""

    username: str
    password: str
    registry_url: str
    expiration: datetime.datetime

    def is_expired(self) -> bool:
        return utcnow() >= self.expiration

    def to_dict(self) -> dict[str, typing.Any]:
        return self.model_dump(mode="json")


class RepositoryPurpose(enum.Enum):
    """Repository purpose enum"""

    Executor = "executor"


class RepositoryTag(enum.Enum):
    """Repository tag enum"""

    CreatedBy = "created_by"
    OrgId = "org_id"
    Purpose = "purpose"  # RepositoryPurpose


class ImageRepository(typing.TypedDict):
    """Image repository details"""

    repository_name: str
    repository_uri: str


class ImageRegistry:
    __roboto_client: RobotoClient

    def __init__(self, roboto_client: RobotoClient) -> None:
        self.__roboto_client = roboto_client

    def create_repository(
        self,
        repository_name: str,
        immutable_image_tags: bool = False,
        org_id: typing.Optional[str] = None,
    ) -> ImageRepository:
        """
        Create a repository for a container image in Roboto's private image registry.
        Images with different tags can be pushed to the same repository.

        Args:
            repository_name: The name of the repository to create.
            immutable_image_tags: Whether to allow image tags to be overwritten. If set to True,
                then any attempt to overwrite an existing image tag will error.

        Returns:
            A dictionary contains the `repository_name` and `repository_uri` of the created repository.
        """
        request_body = CreateImageRepositoryRequest(
            repository_name=repository_name,
            immutable_image_tags=immutable_image_tags,
        )
        response = self.__roboto_client.put(
            "v1/images/repository",
            data=request_body,
            caller_org_id=org_id,
        )
        return response.to_dict(json_path=["data"])

    def delete_image(self, image_uri: str, org_id: typing.Optional[str] = None) -> None:
        """
        Delete a container image from Roboto's private registry.

        Args:
            image_uri: The full URI of the image to delete.
            org_id: ID of organization owning the provided image.
        """
        self.__roboto_client.delete(
            "v1/images/image",
            data=DeleteImageRequest(
                image_uri=image_uri,
            ),
            owner_org_id=org_id,
        )

    def delete_repository(
        self,
        repository_name: str,
        org_id: typing.Optional[str] = None,
        force: bool = False,
    ) -> None:
        """
        Delete a container image from Roboto's private registry.

        Args:
            repository_name: The name of the repository to delete.
            org_id: ID of organization owning the provided image.
            force: Delete all images in the repository before deleting the repository if the repository is not empty.
        """
        self.__roboto_client.delete(
            "v1/images/repository",
            data=DeleteImageRepositoryRequest(
                repository_name=repository_name,
                force=force,
            ),
            owner_org_id=org_id,
        )

    def get_container_image_record(
        self, org_id: str, image_uri: str
    ) -> ContainerImageRecord:
        url_safe_image_uri = urllib.parse.quote_plus(image_uri)
        response = self.__roboto_client.get(
            f"v1/images/image/record/{org_id}/{url_safe_image_uri}",
            owner_org_id=org_id,
        )
        return response.to_record(ContainerImageRecord)

    def get_temporary_credentials(
        self,
        repository_uri: str,
        permissions: Permissions = Permissions.ReadOnly,
        org_id: typing.Optional[str] = None,
    ) -> ContainerCredentials:
        response = self.__roboto_client.get(
            "v1/images/credentials",
            caller_org_id=org_id,
            query={
                "repository_uri": repository_uri,
                "permissions": permissions.value,
            },
        )

        return ContainerCredentials.model_validate(response.to_dict(json_path=["data"]))

    def list_images(
        self,
        repository_name: typing.Optional[str] = None,
        page_token: typing.Optional[str] = None,
        org_id: typing.Optional[str] = None,
    ) -> PaginatedList[ContainerImageRecord]:
        qs_params = dict()
        if repository_name:
            qs_params["repository_name"] = repository_name

        if page_token:
            qs_params["page_token"] = page_token

        response = self.__roboto_client.get(
            "v1/images/image/record/list",
            caller_org_id=org_id,
            query=qs_params,
        )
        return response.to_paginated_list(ContainerImageRecord)

    def list_repositories(
        self,
        page_token: typing.Optional[str] = None,
        org_id: typing.Optional[str] = None,
    ) -> PaginatedList[ContainerImageRepositoryRecord]:
        response = self.__roboto_client.get(
            "v1/images/repository/record/list",
            owner_org_id=org_id,
            query={"page_token": page_token} if page_token else None,
        )
        return response.to_paginated_list(ContainerImageRepositoryRecord)

    def repository_contains_image(
        self,
        repository_name: str,
        image_tag: str,
        org_id: typing.Optional[str] = None,
    ) -> bool:
        urlsafe_repository_name = urllib.parse.quote_plus(repository_name)
        urlsafe_image_tag = urllib.parse.quote_plus(image_tag)

        response = self.__roboto_client.get(
            f"v1/images/repository/{urlsafe_repository_name}/contains/{urlsafe_image_tag}",
            caller_org_id=org_id,
        )
        contains_image = response.to_dict(json_path=["data", "contains_image"])
        return contains_image



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import pathlib

from ...domain.datasets import Dataset
from ..command import (
    ExistingPathlibPath,
    RobotoCommand,
)
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def upload_files(args, context: CLIContext, parser: argparse.ArgumentParser):
    path: pathlib.Path = args.path
    if args.exclude is not None and not path.is_dir():
        parser.error(
            "Exclude filters are only supported for directory uploads, not single files."
        )

    dataset = Dataset.from_id(args.dataset_id, context.roboto_client)

    if path.is_dir():
        dataset.upload_directory(
            directory_path=path,
            exclude_patterns=args.exclude,
        )
    else:
        dataset.upload_files(
            files=[path],
        )


def upload_files_setup_parser(parser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )
    parser.add_argument(
        "-p",
        "--path",
        type=ExistingPathlibPath,
        required=True,
        help="The path to a file or directory to upload.",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        nargs="*",
        help="Zero or more exclude filters (if path points to a directory)",
    )


upload_files_command = RobotoCommand(
    name="upload-files",
    logic=upload_files,
    setup_parser=upload_files_setup_parser,
    command_kwargs={"help": "Uploads a file or directory to a specific dataset."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import json
import typing

import pydantic


def safe_dict_drill(
    target: dict[typing.Any, typing.Any],
    keys: list[typing.Any],
    case_insensitive: bool = False,
) -> typing.Any:
    value = target

    for key in keys:
        if type(value) is not dict:
            return None

        if case_insensitive:
            value = case_insensitive_get(value, key)
        else:
            value = value.get(key, None)

        if value is None:
            return None

    return value


def case_insensitive_get(
    target: dict[str, typing.Any], key: str, default: typing.Any = None
) -> typing.Any:
    matches = list(filter(lambda k: k.lower() == key.lower(), target.keys()))
    n_matches = len(matches)
    if n_matches == 0:
        return default
    elif n_matches == 1:
        return target[matches[0]]
    else:
        raise ValueError(
            f"{n_matches} equivalent case-insensitive keys found for {key}"
        )


def pydantic_jsonable_dict(
    model: pydantic.BaseModel, exclude_none=False, exclude_unset=False
) -> dict:
    return json.loads(
        model.model_dump_json(exclude_none=exclude_none, exclude_unset=exclude_unset)
    )


def pydantic_jsonable_dicts(
    models: collections.abc.Iterable, exclude_none=False, exclude_unset=False
) -> list[dict]:
    return [
        pydantic_jsonable_dict(model, exclude_none, exclude_unset) for model in models
    ]


def pydantic_jsonable_dict_map(
    models: collections.abc.Mapping, exclude_none=False, exclude_unset=False
) -> dict:
    dict_map = {}

    for key, value in models.items():
        dict_map[key] = pydantic_jsonable_dict(value, exclude_none, exclude_unset)

    return dict_map



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic


class CreateDeviceRequest(pydantic.BaseModel):
    """
    Request payload to create a new device
    """

    device_id: str = pydantic.Field(
        description="A user-provided identifier for a device, which is unique within that device's org."
    )
    org_id: typing.Optional[str] = pydantic.Field(
        description="The org to which this device belongs.", default=None
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pathlib
import typing

from ..domain import actions, datasets, orgs
from ..env import RobotoEnv
from ..http import RobotoClient
from .exceptions import ActionRuntimeException
from .file_changeset import (
    FilesChangesetFileManager,
)

_BAD_ENV_BLURB = (
    "This most likely means that you're not running in an action context, "
    + "and are testing a script on a developer machine."
)


class ActionRuntime:
    """
    A utility for performing common lookups and other operations during a Roboto action's runtime. The easiest and
    most common way to initialize this is:

    >>> from roboto import ActionRuntime
    >>> action_runtime = ActionRuntime.from_env()

    ...which will inspect environment variables to initialize the ActionRuntime.

    If you want to test a script using ActionRuntime in a setting such as a developer machine or unit test,
    and you don't want to set environment variables to mirror Roboto's remote execution environment,
    you can also initialize ActionRuntime directly like:

    >>> import pathlib
    >>> from roboto import ActionRuntime
    >>>
    >>> action_runtime = ActionRuntime(
    ...     dataset_id="ds_XXXXXXXXXXXX",
    ...     input_dir=pathlib.Path("/path/to/local/input/dir"),
    ...     invocation_id="iv_XXXXXXXXXXXX",
    ...     org_id="og_XXXXXXXXXXXX",
    ...     output_dir=pathlib.Path("/path/to/local/output/dir"),
    ... )
    """

    __dataset_id: str
    __dataset: typing.Optional[datasets.Dataset] = None
    __file_changeset_manager: typing.Optional[FilesChangesetFileManager] = None
    __input_dir: pathlib.Path
    __invocation_id: str
    __invocation: typing.Optional[actions.Invocation] = None
    __roboto_client: RobotoClient
    __org_id: str
    __org: typing.Optional[orgs.Org] = None
    __output_dir: pathlib.Path

    @classmethod
    def from_env(cls) -> "ActionRuntime":
        """
        Initializes an ActionRuntime based on values in environment variables. This will throw an exception if
        any required environment variables are not available. All of these will be available at runtime in Roboto's
        remote execution environment.

        Example:
            >>> from roboto import ActionRuntime
            >>> action_runtime = ActionRuntime.from_env()
        """
        env = RobotoEnv.default()

        if not env.dataset_id:
            raise ActionRuntimeException(
                "Couldn't find dataset_id from environment. " + _BAD_ENV_BLURB
            )

        if not env.invocation_id:
            raise ActionRuntimeException(
                "Couldn't find invocation_id from environment. " + _BAD_ENV_BLURB
            )

        if not env.org_id:
            raise ActionRuntimeException(
                "Couldn't find org_id from environment. " + _BAD_ENV_BLURB
            )

        if not env.input_dir:
            raise ActionRuntimeException(
                "Couldn't find input_dir from environment. " + _BAD_ENV_BLURB
            )

        input_dir = pathlib.Path(env.input_dir)

        if not input_dir.is_dir():
            raise ActionRuntimeException(
                f"Input directory '{input_dir}' does not exist."
            )

        if not env.output_dir:
            raise ActionRuntimeException(
                "Couldn't find output_dir from environment. " + _BAD_ENV_BLURB
            )

        output_dir = pathlib.Path(env.output_dir)

        if not output_dir.is_dir():
            raise ActionRuntimeException(
                f"Output directory '{output_dir}' does not exist."
            )

        return cls(
            dataset_id=env.dataset_id,
            input_dir=input_dir,
            invocation_id=env.invocation_id,
            org_id=env.org_id,
            output_dir=output_dir,
            roboto_client=RobotoClient.from_env(),
        )

    def __init__(
        self,
        dataset_id: str,
        input_dir: pathlib.Path,
        invocation_id: str,
        org_id: str,
        output_dir: pathlib.Path,
        roboto_client: typing.Optional[RobotoClient] = None,
    ):
        self.__dataset = None
        self.__dataset_id = dataset_id
        self.__file_changeset_manager = None
        self.__input_dir = input_dir
        self.__invocation = None
        self.__invocation_id = invocation_id
        self.__org = None
        self.__org_id = org_id
        self.__output_dir = output_dir
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    @property
    def dataset_id(self) -> str:
        """
        The ID of the dataset whose data this action is operating on.
        """
        return self.__dataset_id

    @property
    def dataset(self) -> datasets.Dataset:
        """
        A :class:`~roboto.domain.datasets.Dataset` object for the dataset whose data this action is operating on.

        This object will be lazy-initialized the first time it is accessed, which might result in a
        :class:`~roboto.exceptions.domain.RobotoNotFoundException` if the dataset does not exist.
        After the first call, the dataset will be cached.

        This is particularly useful for adding tags or metadata to a dataset at runtime, for example:

        >>> from roboto import ActionRuntime
        >>> action_runtime = ActionRuntime.from_env()
        >>> action_runtime.dataset.put_tags(["tagged_by_action"])
        >>> action_runtime.dataset.put_metadata({"voltage_spikes_seen": 693})
        """
        if self.__dataset is None:
            self.__dataset = datasets.Dataset.from_id(
                self.__dataset_id, roboto_client=self.__roboto_client
            )
        return self.__dataset

    @property
    def file_changeset_manager(self) -> FilesChangesetFileManager:
        """
        A :class:`~roboto.action_runtime.file_changeset.FilesChangesetFileManager` which can be used to associate
        tags and metadata with the yet-to-be-uploaded files in this invocation's output directory. In practice, you
        might use this like:

        >>> from roboto import ActionRuntime
        >>> action_runtime = ActionRuntime.from_env()
        >>>
        >>> my_output_file = action_runtime.output_dir / "nested" / "my_output_file.txt"
        >>> my_output_file.write_text("Hello World")
        >>>
        >>> action_runtime.file_changeset_manager.put_tags("nested/my_output_file.txt", ["tagged_by_action"])
        >>> action_runtime.file_changeset_manager.put_fields(
        ... "nested/my_output_file.txt", {"roboto_proficiency": "extreme - I can annotate output files!"}
        ... )

        This only works for files that don't exist yet. To tag existing files (such as files in the input directory),
        you should instead use:

        >>> from roboto import ActionRuntime
        >>> action_runtime = ActionRuntime.from_env()
        >>>
        >>> existing_file = action_runtime.dataset.get_file_by_path("some_file_that_already_exists.txt")
        >>> existing_file.put_tags(["tagged_by_action"])
        >>> existing_file.put_metadata({"roboto_proficiency": "also extreme - I can annotate input files!"})

        For more info, see the top-level docs on the FilesChangesetFileManager class.
        """
        if self.__file_changeset_manager is None:
            self.__file_changeset_manager = FilesChangesetFileManager()
        return self.__file_changeset_manager

    @property
    def input_dir(self) -> pathlib.Path:
        """
        The directory where the action's input files are located.
        """
        return self.__input_dir

    @property
    def invocation_id(self) -> str:
        """
        The ID of the currently running action invocation.
        """
        return self.__invocation_id

    @property
    def invocation(self) -> actions.Invocation:
        """
        An :class:`~roboto.domain.actions.Invocation` object for the currently running action invocation.

        This object will be lazy-initialized the first time it is accessed, which might result in a
        :class:`~roboto.exceptions.domain.RobotoNotFoundException` if the invocation does not exist.
        After the first call, the invocation will be cached.
        """
        if self.__invocation is None:
            self.__invocation = actions.Invocation.from_id(
                self.__invocation_id, roboto_client=self.__roboto_client
            )
        return self.__invocation

    @property
    def org_id(self) -> str:
        """
        The ID of the org which invoked the currently running action.
        """
        return self.__org_id

    @property
    def org(self) -> orgs.Org:
        """
        An :class:`~roboto.domain.orgs.Org` object for the org which invoked the currently running action.

        This object will be lazy-initialized the first time it is accessed, which might result in a
        :class:`~roboto.exceptions.domain.RobotoNotFoundException` if the org does not exist.
        After the first call, the org will be cached.
        """
        if self.__org is None:
            self.__org = orgs.Org.from_id(
                self.__org_id, roboto_client=self.__roboto_client
            )
        return self.__org

    @property
    def output_dir(self) -> pathlib.Path:
        """
        The directory where the action's output files are expected. After the user portion of the action runtime
        concludes (i.e. when their container exits with a 0 exit code), every file in this directory will be uploaded
        to the dataset associated with this action invocation.
        """
        return self.__output_dir



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .event import Event
from .operations import (
    CreateEventRequest,
    QueryEventsForAssociationsRequest,
    UpdateEventRequest,
)
from .record import EventRecord

__all__ = [
    "CreateEventRequest",
    "Event",
    "EventRecord",
    "QueryEventsForAssociationsRequest",
    "UpdateEventRequest",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json
import typing

from ...domain import actions
from ...query import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
    QuerySpecification,
    SortDirection,
)
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..common_args import add_org_arg
from ..context import CLIContext


def search(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    conditions: list[typing.Union[Condition, ConditionGroup]] = []
    if args.name:
        conditions.append(
            Condition(
                field="name",
                comparator=Comparator.Equals,
                value=args.name,
            )
        )

    if args.metadata:
        for key, value in args.metadata.items():
            conditions.append(
                Condition(
                    field=f"metadata.{key}",
                    comparator=Comparator.Equals,
                    value=value,
                )
            )

    if args.tag:
        for tag in args.tag:
            conditions.append(
                Condition(
                    field="tags",
                    comparator=Comparator.Contains,
                    value=tag,
                )
            )

    query_args: dict[str, typing.Any] = {
        "sort_direction": SortDirection.Descending,
    }
    if conditions:
        if len(conditions) == 1:
            query_args["condition"] = conditions[0]
        else:
            query_args["condition"] = ConditionGroup(
                conditions=conditions, operator=ConditionOperator.And
            )

    query = QuerySpecification(**query_args)
    accessibility = (
        actions.Accessibility.ActionHub
        if args.actionhub
        else actions.Accessibility.Organization
    )
    matching_actions = actions.Action.query(
        query,
        accessibility=accessibility,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    print(json.dumps([action.to_dict() for action in matching_actions], indent=2))


def search_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--name",
        required=False,
        action="store",
        help="If querying by action name, must provide an exact match; patterns are not accepted.",
    )
    parser.add_argument(
        "--actionhub",
        required=False,
        action="store_true",
        help="Search Actions published to the Roboto Action Hub",
    )
    parser.add_argument(
        "--metadata",
        required=False,
        metavar="KEY=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``key=value`` pairs which represent action metadata. "
            "``value`` is parsed as JSON. E.g.: ``--metadata foo=bar --metadata baz.nested=200``"
        ),
    )
    parser.add_argument(
        "--tag",
        required=False,
        type=str,
        nargs="*",
        help="One or more tags associated with this action. E.g.: ``--tag foo --tag bar``",
        action="extend",
    )
    add_org_arg(parser=parser)


search_command = RobotoCommand(
    name="search",
    logic=search,
    setup_parser=search_parser,
    command_kwargs={"help": "Search for existing actions."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import typing

import pydantic
from pydantic import model_validator

from ...association import Association
from ...sentinels import (
    NotSet,
    NotSetType,
    is_set,
)
from ...updates import (
    MetadataChangeset,
    TaglessMetadataChangeset,
)
from .record import (
    CanonicalDataType,
    MessagePathRecord,
    RepresentationRecord,
    RepresentationStorageFormat,
)


class BaseAddRepresentationRequest(pydantic.BaseModel):
    """Request payload to add a representation to a topic"""

    association: Association
    storage_format: RepresentationStorageFormat
    version: int


class SetDefaultRepresentationRequest(BaseAddRepresentationRequest):
    """Memorialize a representation of topic data contained within a source recording file"""

    # 'ignore' used to avoid backwards incompatible change to remove `org_id` from BaseAddRepresentationRequest.
    # Should be changed back to 'forbid' for SDK v1.0
    model_config = pydantic.ConfigDict(extra="ignore")


class AddMessagePathRepresentationRequest(BaseAddRepresentationRequest):
    """Associate a MessagePath with a Representation"""

    message_path_id: str

    # 'ignore' used to avoid backwards incompatible change to remove `org_id` from BaseAddRepresentationRequest
    # Should be changed back to 'forbid' for SDK v1.0
    model_config = pydantic.ConfigDict(extra="ignore")


class MessagePathRepresentationMapping(pydantic.BaseModel):
    """Latest representation of topic data in particular storage format containing given message paths"""

    message_paths: collections.abc.MutableSequence[MessagePathRecord]
    representation: RepresentationRecord


class AddMessagePathRequest(pydantic.BaseModel):
    """Associate a MessagePath with a Topic."""

    message_path: str
    data_type: str
    canonical_data_type: CanonicalDataType
    metadata: dict[str, typing.Any] = pydantic.Field(
        default_factory=dict,
        description="Initial key-value pairs to associate with this topic message path for discovery and search, e.g. "
        + "`{ 'min': 0.71, 'max': 1.77, 'classification': 'my-custom-classification-tag' }`",
    )

    model_config = pydantic.ConfigDict(extra="forbid")


class UpdateMessagePathRequest(pydantic.BaseModel):
    """Update attributes of a message path within a topic"""

    message_path: str
    """Message path name (required)."""

    metadata_changeset: typing.Union[TaglessMetadataChangeset, NotSetType] = NotSet
    """A set of changes to the message path's metadata (optional)."""

    data_type: typing.Union[str, NotSetType] = NotSet
    """Native data type for the data under this message path (optional)."""

    canonical_data_type: typing.Union[CanonicalDataType, NotSetType] = NotSet
    """Canonical Roboto data type for the data under this message path (optional).

    Note: updating this attribute should be done with care, as it affects Roboto's
    ability to interpret and visualize the data.
    """

    model_config = pydantic.ConfigDict(extra="forbid")

    def has_updates(self) -> bool:
        """Checks whether this request would result in any message path modifications."""

        return (
            is_set(self.data_type)
            or is_set(self.canonical_data_type)
            or (
                is_set(self.metadata_changeset)
                and self.metadata_changeset.has_changes()
            )
        )


class DeleteMessagePathRequest(pydantic.BaseModel):
    """Delete a message path from a topic."""

    message_path: str
    """Message path name."""

    model_config = pydantic.ConfigDict(extra="forbid")


class MessagePathChangeset(pydantic.BaseModel):
    """A set of changes to add, delete or update message paths on a topic."""

    message_paths_to_add: collections.abc.Sequence[AddMessagePathRequest] | None = None
    """Message paths to add to a topic."""

    message_paths_to_delete: (
        collections.abc.Sequence[DeleteMessagePathRequest] | None
    ) = None
    """Message paths to delete from a topic."""

    message_paths_to_update: (
        collections.abc.Sequence[UpdateMessagePathRequest] | None
    ) = None
    """Message paths to update on a topic."""

    replace_all: bool = False
    """Flag indicating whether this changeset should replace all message paths on a topic.

    It assumes that the replacement message paths will be provided via ``message_paths_to_add``. Rather than
    setting this flag directly, use appropriate class methods such as ``from_replacement_message_paths``.
    """

    @model_validator(mode="after")
    def check_replace_all_correctness(self) -> MessagePathChangeset:
        if self.replace_all and (
            self.message_paths_to_add is None
            or not (
                self.message_paths_to_update is None
                and self.message_paths_to_delete is None
            )
        ):
            raise ValueError("replace_all must only be used with message_paths_to_add")

        return self

    @classmethod
    def from_replacement_message_paths(
        cls, message_paths: collections.abc.Sequence[AddMessagePathRequest]
    ) -> MessagePathChangeset:
        """A changeset that replaces any existing message paths with the ones provided."""

        return cls(message_paths_to_add=message_paths, replace_all=True)

    def has_changes(self) -> bool:
        """Checks whether the changeset contains any actual changes."""

        return self.replace_all or not (
            (self.message_paths_to_add is None or not self.message_paths_to_add)
            and (
                self.message_paths_to_update is None or not self.message_paths_to_update
            )
            and (
                self.message_paths_to_delete is None or not self.message_paths_to_delete
            )
        )


class CreateTopicRequest(pydantic.BaseModel):
    """Memorialize a Topic contained within a source recording file"""

    # Required
    association: Association
    topic_name: str

    # Optional
    end_time: typing.Optional[int] = None
    message_count: typing.Optional[int] = None
    metadata: typing.Optional[collections.abc.Mapping[str, typing.Any]] = None
    schema_checksum: typing.Optional[str] = None
    schema_name: typing.Optional[str] = None
    start_time: typing.Optional[int] = None
    message_paths: typing.Optional[collections.abc.Sequence[AddMessagePathRequest]] = (
        None
    )

    # 'ignore' used to avoid backwards incompatible change to remove `org_id`
    # Should be changed back to 'forbid' for SDK v1.0
    model_config = pydantic.ConfigDict(extra="ignore")


class UpdateTopicRequest(pydantic.BaseModel):
    """Request payload to update a Topic"""

    end_time: typing.Union[typing.Optional[int], NotSetType] = NotSet
    message_count: typing.Union[int, NotSetType] = NotSet
    schema_checksum: typing.Union[typing.Optional[str], NotSetType] = NotSet
    schema_name: typing.Union[typing.Optional[str], NotSetType] = NotSet
    start_time: typing.Union[typing.Optional[int], NotSetType] = NotSet
    metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet
    message_path_changeset: typing.Union[MessagePathChangeset, NotSetType] = NotSet

    model_config = pydantic.ConfigDict(
        extra="forbid", json_schema_extra=NotSetType.openapi_schema_modifier
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .message_path import MessagePath
from .operations import (
    AddMessagePathRepresentationRequest,
    AddMessagePathRequest,
    CreateTopicRequest,
    DeleteMessagePathRequest,
    MessagePathChangeset,
    MessagePathRepresentationMapping,
    SetDefaultRepresentationRequest,
    UpdateMessagePathRequest,
    UpdateTopicRequest,
)
from .record import (
    CanonicalDataType,
    MessagePathRecord,
    MessagePathStatistic,
    RepresentationRecord,
    RepresentationStorageFormat,
    TopicRecord,
)
from .topic import Topic
from .topic_data_service import TopicDataService

__all__ = (
    "AddMessagePathRequest",
    "AddMessagePathRepresentationRequest",
    "CreateTopicRequest",
    "CanonicalDataType",
    "DeleteMessagePathRequest",
    "MessagePath",
    "MessagePathChangeset",
    "MessagePathRepresentationMapping",
    "MessagePathRecord",
    "MessagePathStatistic",
    "RepresentationRecord",
    "RepresentationStorageFormat",
    "SetDefaultRepresentationRequest",
    "Topic",
    "TopicDataService",
    "TopicRecord",
    "UpdateMessagePathRequest",
    "UpdateTopicRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import pathlib
import typing
import urllib.parse

import boto3
import botocore.config
import botocore.credentials
import botocore.session

from ...association import Association
from ...http import BatchRequest, RobotoClient
from ...query import QuerySpecification
from ...sentinels import (
    NotSet,
    NotSetType,
    remove_not_set,
)
from ...updates import MetadataChangeset
from ..topics import Topic
from .operations import (
    ImportFileRequest,
    RenameFileRequest,
    UpdateFileRecordRequest,
)
from .progress import (
    NoopProgressMonitorFactory,
    ProgressMonitorFactory,
)
from .record import CredentialProvider, FileRecord


class File:
    """
    Files can be uploaded to datasets. Once uploaded, they can be tagged,
    post-processed by actions, added to collections, visualized, and searched.
    """

    __record: FileRecord
    __roboto_client: RobotoClient

    @staticmethod
    def construct_s3_obj_arn(bucket: str, key: str, partition: str = "aws") -> str:
        return f"arn:{partition}:s3:::{bucket}/{key}"

    @staticmethod
    def construct_s3_obj_uri(
        bucket: str, key: str, version: typing.Optional[str] = None
    ) -> str:
        base_uri = f"s3://{bucket}/{key}"
        if version:
            base_uri += f"?versionId={version}"
        return base_uri

    @staticmethod
    def generate_s3_client(
        credential_provider: CredentialProvider, tcp_keepalive: bool = True
    ):
        creds = credential_provider()
        refreshable_credentials = (
            botocore.credentials.RefreshableCredentials.create_from_metadata(
                metadata=creds,
                refresh_using=credential_provider,
                method="roboto-api",
            )
        )
        botocore_session = botocore.session.get_session()
        botocore_session._credentials = refreshable_credentials
        botocore_session.set_config_variable("region", creds["region"])
        session = boto3.Session(botocore_session=botocore_session)

        return session.client(
            "s3", config=botocore.config.Config(tcp_keepalive=tcp_keepalive)
        )

    @classmethod
    def from_id(
        cls,
        file_id: str,
        version_id: typing.Optional[int] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "File":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(
            f"v1/files/record/{file_id}",
            query={"version_id": version_id} if version_id is not None else None,
        ).to_record(FileRecord)
        return cls(record, roboto_client)

    @classmethod
    def from_path_and_dataset_id(
        cls,
        file_path: typing.Union[str, pathlib.Path],
        dataset_id: str,
        version_id: typing.Optional[int] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "File":
        roboto_client = RobotoClient.defaulted(roboto_client)
        url_quoted_file_path = urllib.parse.quote(str(file_path), safe="")
        record = roboto_client.get(
            f"v1/files/record/path/{url_quoted_file_path}/association/{dataset_id}",
            query={"version_id": version_id} if version_id is not None else None,
        ).to_record(FileRecord)
        return cls(record, roboto_client)

    @classmethod
    def import_batch(
        cls,
        requests: collections.abc.Sequence[ImportFileRequest],
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Sequence["File"]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        # Requests explicitly need to be cast to list because of Pydantic serialization not working appropriately
        # with collections.abc
        request: BatchRequest[ImportFileRequest] = BatchRequest(requests=list(requests))

        records = roboto_client.post(
            "v1/files/import/batch", data=request, idempotent=True
        ).to_record_list(FileRecord)
        return [cls(record, roboto_client) for record in records]

    @classmethod
    def query(
        cls,
        spec: typing.Optional[QuerySpecification] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
        owner_org_id: typing.Optional[str] = None,
    ) -> collections.abc.Generator["File", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        spec = spec or QuerySpecification()

        known = set(FileRecord.model_fields.keys())
        actual = set()
        for field in spec.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of File"
                if plural
                else "is not a known attribute of File"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        while True:
            paginated_results = roboto_client.post(
                "v1/files/query", data=spec, owner_org_id=owner_org_id, idempotent=True
            ).to_paginated_list(FileRecord)

            for record in paginated_results.items:
                yield cls(record, roboto_client)

            if paginated_results.next_token:
                spec.after = paginated_results.next_token
            else:
                break

    def __init__(
        self, record: FileRecord, roboto_client: typing.Optional[RobotoClient] = None
    ):
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__record = record

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def dataset_id(self) -> str:
        return self.__record.association_id

    @property
    def description(self) -> typing.Optional[str]:
        return self.__record.description

    @property
    def file_id(self) -> str:
        return self.__record.file_id

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def record(self) -> FileRecord:
        return self.__record

    @property
    def relative_path(self) -> str:
        return self.__record.relative_path

    @property
    def metadata(self) -> dict[str, typing.Any]:
        return self.__record.metadata

    @property
    def tags(self) -> list[str]:
        return self.__record.tags

    @property
    def uri(self) -> str:
        return self.__record.uri

    @property
    def version(self) -> int:
        return self.__record.version

    def delete(self) -> None:
        self.__roboto_client.delete(f"/v1/files/{self.file_id}")

    def download(
        self,
        local_path: pathlib.Path,
        credential_provider: CredentialProvider,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ):
        local_path.parent.mkdir(parents=True, exist_ok=True)
        s3_client = File.generate_s3_client(credential_provider)

        res = s3_client.head_object(Bucket=self.record.bucket, Key=self.record.key)
        download_bytes = int(res.get("ContentLength", 0))

        source = self.record.key.replace(f"{self.record.org_id}/datasets/", "")

        progress_monitor = progress_monitor_factory.download_monitor(
            source=source, size=download_bytes
        )
        try:
            s3_client.download_file(
                Bucket=self.record.bucket,
                Key=self.record.key,
                Filename=str(local_path),
                Callback=progress_monitor.update,
            )
        finally:
            progress_monitor.close()

    def get_signed_url(
        self,
        override_content_type: typing.Optional[str] = None,
        override_content_disposition: typing.Optional[str] = None,
    ) -> str:
        query_params: dict[str, str] = {}

        if override_content_disposition:
            query_params["override_content_disposition"] = override_content_disposition

        if override_content_type:
            query_params["override_content_type"] = override_content_type

        res = self.__roboto_client.get(
            f"v1/files/{self.file_id}/signed-url",
            query=query_params,
            owner_org_id=self.org_id,
        )
        return res.to_dict(json_path=["data", "url"])

    def get_topic(self, topic_name: str) -> Topic:
        return Topic.from_name_and_file(
            topic_name=topic_name,
            file_id=self.file_id,
            owner_org_id=self.org_id,
            roboto_client=self.__roboto_client,
        )

    def get_topics(
        self,
        include: typing.Optional[collections.abc.Sequence[str]] = None,
        exclude: typing.Optional[collections.abc.Sequence[str]] = None,
    ) -> collections.abc.Generator["Topic", None, None]:
        for topic in Topic.get_by_file(
            owner_org_id=self.org_id,
            file_id=self.file_id,
            roboto_client=self.__roboto_client,
        ):
            if include is not None and topic.name not in include:
                continue

            if exclude is not None and topic.name in exclude:
                continue

            yield topic

    def put_metadata(self, metadata: dict[str, typing.Any]) -> "File":
        return self.update(metadata_changeset=MetadataChangeset(put_fields=metadata))

    def put_tags(self, tags: list[str]) -> "File":
        return self.update(metadata_changeset=MetadataChangeset(put_tags=tags))

    def refresh(self) -> "File":
        self.__record = self.__roboto_client.get(
            f"v1/files/record/{self.file_id}"
        ).to_record(FileRecord)
        return self

    def rename_file(self, file_id: str, new_path: str) -> FileRecord:

        response = self.__roboto_client.put(
            f"v1/files/{self.file_id}/rename",
            data=RenameFileRequest(
                association_id=self.dataset_id,
                new_path=new_path,
            ),
        )

        return response.to_record(FileRecord)

    def to_association(self) -> Association:
        return Association.file(self.file_id, self.version)

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")

    def update(
        self,
        description: typing.Optional[typing.Union[str, NotSetType]] = NotSet,
        metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet,
    ) -> "File":
        request = remove_not_set(
            UpdateFileRecordRequest(
                description=description, metadata_changeset=metadata_changeset
            )
        )
        self.__record = self.__roboto_client.put(
            f"v1/files/record/{self.file_id}", data=request
        ).to_record(FileRecord)
        return self



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pydantic

from .record import CommentEntityType


class CreateCommentRequest(pydantic.BaseModel):
    """
    Request payload to create a comment
    """

    entity_type: CommentEntityType
    entity_id: str
    comment_text: str


class UpdateCommentRequest(pydantic.BaseModel):
    """
    Request payload to update a comment
    """

    comment_text: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing

from ...exceptions import RobotoDomainException
from ...http import RobotoClient
from ..tokens import (
    CreateTokenRequest,
    Token,
    TokenRecord,
)
from .operations import CreateDeviceRequest
from .record import DeviceRecord


class Device:
    """
    A device is a non-human entity which can interact with Roboto on behalf of a specific organization.
    Each device is identified by a device_id, which is unique among all devices in the organization to which it belongs.

    The most typical device is a robot which uploads its log data to Roboto, either directly from its on-board software
    stack, or indirectly through an automatic upload station or human-in-the-loop upload process. Its device ID
    may be a string that represents its serial number in a scheme native to its organization.

    A dedicated uploader station which connects to many different robots and uploads data on their behalf could also
    be modeled as a device.

    API access tokens can be allocated for devices, and these tokens can be used to authenticate Roboto requests made
    on behalf of a device.
    """

    __roboto_client: RobotoClient
    __record: DeviceRecord

    # Class methods + Constructor

    @classmethod
    def create(
        cls,
        device_id: str,
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Device":
        """
        Registers a device with Roboto.
        """
        roboto_client = RobotoClient.defaulted(roboto_client)
        request = CreateDeviceRequest(device_id=device_id, org_id=caller_org_id)
        record = roboto_client.post(
            "v1/devices/create", caller_org_id=caller_org_id, data=request
        ).to_record(DeviceRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def for_org(
        cls, org_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> collections.abc.Generator["Device", None, None]:
        """
        List all devices registered for a given org.

        Args:
            org_id: The org to list devices for
            roboto_client: Common parameters required to construct any Device object

        Returns:
            A generator of Device objects. For orgs with a large number of devices, this may involve multiple service
            calls, and the generator will yield results as they become available.
        """
        roboto_client = RobotoClient.defaulted(roboto_client)

        next_token: typing.Optional[str] = None
        while True:
            query_params: dict[str, typing.Any] = {}
            if next_token:
                query_params["page_token"] = str(next_token)

            results = roboto_client.get(
                f"v1/devices/org/{org_id}",
                query=query_params,
            ).to_paginated_list(DeviceRecord)

            for item in results.items:
                yield cls(record=item, roboto_client=roboto_client)

            next_token = results.next_token
            if not next_token:
                break

    @classmethod
    def from_id(
        cls,
        device_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
        org_id: typing.Optional[str] = None,
    ) -> "Device":
        """
        Args:
            device_id: The device ID to look up. See :func:`Device.device_id` for more details.
            roboto_client: Common parameters required to construct any Device object
            org_id: The org to which the device belongs.
                If not specified by a caller who only belongs to one org, will default to the org_id of that org.
                If not specified by a caller who belongs to multiple orgs, will raise an exception.

        Returns:
            A Device object representing the specified device

        Raises:
            RobotoNotFoundException: If the specified device is not registered with Roboto
        """
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(
            f"v1/devices/id/{device_id}",
            owner_org_id=org_id,
        ).to_record(DeviceRecord)
        return cls(record=record, roboto_client=roboto_client)

    def __init__(
        self, record: DeviceRecord, roboto_client: typing.Optional[RobotoClient] = None
    ):
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__record = record

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def device_id(self) -> str:
        """
        This device's ID. Device ID is a user-provided identifier for a device, which is unique within the
        device's org.
        """
        return self.__record.device_id

    @property
    def org_id(self) -> str:
        """
        The ID of the org to which this device belongs.
        """
        return self.__record.org_id

    @property
    def record(self) -> DeviceRecord:
        """
        The underlying DeviceRecord object which represents this device. This is often used as the wire representation
        of a device during API requests, and is subject to evolve over time. You should not program against this
        if avoidable.
        """
        return self.__record

    def create_token(
        self,
        expiry_days: int = 366,
        name: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
    ) -> tuple[Token, str]:
        request = CreateTokenRequest(
            expiry_days=expiry_days,
            name=name or f"{self.org_id}_{self.device_id}",
            description=description,
        )

        record = self.__roboto_client.post(
            f"v1/devices/id/{self.device_id}/tokens",
            owner_org_id=self.org_id,
            data=request,
        ).to_record(TokenRecord)

        if record.secret is None:
            raise RobotoDomainException(
                "Token was generated without returning secret value, this should never happen. "
                + "Please reach out to support@roboto.ai"
            )

        return (
            Token(record=record, roboto_client=self.__roboto_client),
            record.secret,
        )

    def delete(self) -> None:
        """
        Deletes this device.
        """
        self.__roboto_client.delete(
            f"v1/devices/id/{self.device_id}", owner_org_id=self.org_id
        )

    def tokens(self) -> collections.abc.Sequence[Token]:
        records = self.__roboto_client.get(
            f"v1/devices/id/{self.device_id}/tokens", owner_org_id=self.org_id
        ).to_record_list(TokenRecord)
        return [
            Token(record=record, roboto_client=self.__roboto_client)
            for record in records
        ]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import os
from typing import Optional

import boto3
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import (
    ReadOnlyCredentials,
)

from ..exceptions import RobotoDeprecatedException
from ..logging import default_logger
from .request import HttpRequest

logger = default_logger()


class BearerTokenDecorator:
    """
    Decorates requests with a static, unchanging bearer token.
    """

    __auth_header: str

    def __init__(self, token: str):
        if token.startswith("robo_pat"):
            raise RobotoDeprecatedException(
                "You're using an auth token created before March 20th, 2024. These are being phased out in favor of a "
                + "newer token format. Please create a new token via https://app.roboto.ai/settings?tab=2."
            )

        self.__auth_header = f"Bearer {token}"

    def __call__(self, request: HttpRequest) -> HttpRequest:
        if request.headers is None:
            request.headers = {}

        request.headers["Authorization"] = self.__auth_header
        return request


class SigV4AuthDecorator:
    __credentials: ReadOnlyCredentials
    __region: str
    __service: str

    @staticmethod
    def lookup_credentials() -> ReadOnlyCredentials:
        session = boto3.Session()
        creds = session.get_credentials()
        if creds is None:
            raise RuntimeError("No AWS credentials found")

        return creds.get_frozen_credentials()

    @staticmethod
    def lookup_region() -> str:
        session = boto3.Session()
        if session.region_name:
            return session.region_name

        region = os.environ.get("AWS_REGION", os.environ.get("AWS_DEFAULT_REGION"))
        if not region:
            raise ValueError("One of AWS_REGION or AWS_DEFAULT_REGION must be set.")
        return region

    def __init__(
        self,
        service: str = "execute-api",
        credentials: Optional[ReadOnlyCredentials] = None,
        region: Optional[str] = None,
    ):
        self.__credentials = (
            credentials if credentials else SigV4AuthDecorator.lookup_credentials()
        )
        self.__region = region if region else SigV4AuthDecorator.lookup_region()
        self.__service = service

    def __call__(self, request: HttpRequest) -> HttpRequest:
        if "Host" not in request.headers:
            request.append_headers({"Host": request.hostname})

        aws_request = AWSRequest(
            method=request.method.upper(), url=request.url, data=request.body
        )
        aws_request.context["payload_signing_enabled"] = True
        SigV4Auth(self.__credentials, self.__service, self.__region).add_auth(
            aws_request
        )
        request.append_headers(dict(aws_request.headers.items()))
        return request



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing

import pathspec

from .query import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
)


def path_to_pattern(path: str) -> str:
    """
    Transform a literal path into a Git wildmatch pattern that matches that literal path.
    """
    escaped = path.replace("[", "\\[")
    escaped = escaped.replace("]", "\\]")
    return escaped


def excludespec_from_patterns(
    exclude_patterns: typing.Optional[collections.abc.Iterable[str]] = None,
) -> typing.Optional[pathspec.PathSpec]:
    """
    Transform a list of Git wildmatch patterns into a pathspec.PathSpec.
    """
    return (
        None
        if exclude_patterns is None
        else pathspec.GitIgnoreSpec.from_lines(exclude_patterns)
    )


def pathspec_from_patterns(
    patterns: collections.abc.Iterable[str],
) -> pathspec.PathSpec:
    """
    Transform a list of Git wildmatch patterns into a pathspec.PathSpec.
    """
    return pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, patterns)


def path_matches(
    path: str,
    include_patterns: typing.Optional[collections.abc.Iterable[str]] = None,
    exclude_patterns: typing.Optional[collections.abc.Iterable[str]] = None,
) -> bool:
    """
    Does the given path match the given include pattern(s) and not match the given exclude pattern(s)?
    """
    # Include patterns are provided, and the file isn't included, ignore it
    if include_patterns is not None:
        if not pathspec_from_patterns(include_patterns).match_file(path):
            return False

    # Exclude pattern is provided, and the file is included, ignore it
    if exclude_patterns is not None:
        if pathspec_from_patterns(exclude_patterns).match_file(path):
            return False

    return True


def pattern_to_like_value(pattern: str) -> str:
    return pattern.replace("**/*", "%").replace("**", "%").replace("*", "%")


def patterns_to_condition_group(
    patterns: collections.abc.Iterable[str],
    path_field_name: str,
    comparator: Comparator,
    operator: ConditionOperator,
) -> ConditionGroup:
    """
    Transform a list of Git wildmatch patterns into a ConditionGroup.
    """
    return ConditionGroup(
        conditions=[
            Condition(
                field=path_field_name,
                comparator=comparator,
                value=pattern_to_like_value(pattern),
            )
            for pattern in patterns
        ],
        operator=operator,
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...image_registry import ImageRegistry
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def delete_repository(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    image_registry = ImageRegistry(context.roboto_client)
    image_registry.delete_repository(
        args.repository_name, org_id=args.org, force=args.force
    )


def delete_repository_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "repository_name",
        action="store",
        help="The name of the image repository to delete.",
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help="Force deletion of the repository, even if it is not empty.",
    )


delete_repo_command = RobotoCommand(
    name="delete-repository",
    logic=delete_repository,
    setup_parser=delete_repository_parser,
    command_kwargs={
        "help": (
            "Delete an image repository hosted in Roboto's image registry. "
            "Requires Docker CLI."
        )
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import atexit
import collections.abc
import concurrent.futures
import contextlib
import pathlib
import time
import typing
import urllib.request

from ...association import AssociationType
from ...http import RobotoClient
from ...logging import default_logger
from ...time import Time, to_epoch_nanoseconds
from .mcap_reader import McapReader
from .operations import (
    MessagePathRepresentationMapping,
)
from .record import (
    MessagePathRecord,
    RepresentationRecord,
    RepresentationStorageFormat,
)

logger = default_logger()


OUTFILE_NAME_PATTERN = "{repr_id}_{file_id}.mcap"


class TopicDataService:
    """
    Data service that extracts common logic for retrieving topic data ingested by Roboto.
    This is not intended as a public API. To access topic data,
    prefer :py:meth:`~roboto.domain.topics.Topic.get_data` instead.
    """

    DEFAULT_CACHE_DIR: typing.ClassVar[pathlib.Path] = (
        pathlib.Path.home() / ".cache" / "roboto" / "topic-data"
    )
    LOG_TIME_ATTR_NAME: typing.ClassVar[str] = "log_time"

    __cache_dir: pathlib.Path
    __roboto_client: RobotoClient

    @staticmethod
    def garbage_collect_old_topic_data(
        cache_dir: pathlib.Path,
        expire_after_ns: int = 7
        * 24
        * 60
        * 60
        * 100
        * 1_000_000,  # 7 days in nanoseconds
    ):
        pattern_as_glob = OUTFILE_NAME_PATTERN.format(file_id="*", repr_id="*")
        now_in_nanoseconds = time.time_ns()
        for mcap_file in cache_dir.glob(f"**/{pattern_as_glob}"):
            if (now_in_nanoseconds - mcap_file.stat().st_atime_ns) > expire_after_ns:
                mcap_file.unlink()

    def __init__(
        self,
        roboto_client: RobotoClient,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ):
        self.__roboto_client = roboto_client
        self.__cache_dir = (
            pathlib.Path(cache_dir)
            if cache_dir is not None
            else TopicDataService.DEFAULT_CACHE_DIR
        )

    def get_data(
        self,
        topic_id: str,
        message_paths_include: typing.Optional[collections.abc.Sequence[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Sequence[str]] = None,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir_override: typing.Union[str, pathlib.Path, None] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        cache_dir = (
            pathlib.Path(cache_dir_override)
            if cache_dir_override is not None
            else self.__cache_dir
        )
        if not cache_dir.exists():
            cache_dir.mkdir(parents=True)

        message_path_repr_mappings_response = self.__roboto_client.get(
            f"v1/topics/id/{topic_id}/message-path/representations/{RepresentationStorageFormat.MCAP.value}"
        )
        message_path_repr_mappings = message_path_repr_mappings_response.to_record_list(
            MessagePathRepresentationMapping
        )

        # Exclude a MessagePathRepresentationMapping if no message paths remain after applying message path filters.
        message_path_repr_mappings = [
            message_path_repr_map.model_copy(
                update={
                    "message_paths": self.__filter_message_paths(
                        message_path_repr_map.message_paths,
                        include_paths=message_paths_include,
                        exclude_paths=message_paths_exclude,
                    )
                }
            )
            for message_path_repr_map in message_path_repr_mappings
        ]
        message_path_repr_mappings = [
            message_path_repr_map
            for message_path_repr_map in message_path_repr_mappings
            if message_path_repr_map.message_paths
        ]

        repr_id_to_outfile_map: dict[str, pathlib.Path] = {}
        download_list: list[MessagePathRepresentationMapping] = []
        for message_path_repr_map in message_path_repr_mappings:
            representation = message_path_repr_map.representation
            outfile = self.__representation_out_file(
                representation, cache_dir=cache_dir
            )
            repr_id_to_outfile_map[representation.representation_id] = outfile
            if not outfile.exists():
                association = representation.association
                if association.association_type != AssociationType.File:
                    logger.warning(
                        "Unable to get data for message paths %r",
                        [
                            record.message_path
                            for record in message_path_repr_map.message_paths
                        ],
                    )
                    continue
                download_list.append(message_path_repr_map)

        if download_list:
            self.__download_representations(
                message_path_repr_mappings=download_list,
                repr_id_to_outfile_map=repr_id_to_outfile_map,
                cache_dir=cache_dir,
            )

        normalized_start_time = (
            to_epoch_nanoseconds(start_time) if start_time is not None else None
        )
        normalized_end_time = (
            to_epoch_nanoseconds(end_time) if end_time is not None else None
        )

        with contextlib.ExitStack() as exit_stack:
            readers = [
                McapReader(
                    stream=exit_stack.enter_context(
                        repr_id_to_outfile_map[
                            message_path_repr_map.representation.representation_id
                        ].open(mode="rb")
                    ),
                    message_paths=message_path_repr_map.message_paths,
                    start_time=normalized_start_time,
                    end_time=normalized_end_time,
                )
                for message_path_repr_map in message_path_repr_mappings
            ]
            while any(reader.has_next for reader in readers):
                full_record = {}
                next_earliest_timestamp = min(
                    reader.next_timestamp for reader in readers
                )
                full_record[TopicDataService.LOG_TIME_ATTR_NAME] = (
                    next_earliest_timestamp
                )
                for reader in readers:
                    if reader.next_message_is_time_aligned(next_earliest_timestamp):
                        decoded_message = reader.next()
                        if decoded_message is None:
                            continue
                        full_record.update(decoded_message.to_dict())

                yield full_record

    def __download_representations(
        self,
        message_path_repr_mappings: collections.abc.Sequence[
            MessagePathRepresentationMapping
        ],
        repr_id_to_outfile_map: collections.abc.Mapping[str, pathlib.Path],
        cache_dir: pathlib.Path,
    ) -> None:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future_to_representation_mapping = {
                executor.submit(
                    self.__download_roboto_file,
                    message_path_repr_map.representation,
                    repr_id_to_outfile_map[
                        message_path_repr_map.representation.representation_id
                    ],
                ): message_path_repr_map
                for message_path_repr_map in message_path_repr_mappings
            }
            for future in concurrent.futures.as_completed(
                future_to_representation_mapping
            ):
                message_path_repr_map = future_to_representation_mapping[future]
                try:
                    future.result()
                except Exception:
                    logger.exception(
                        "Unable to get data for message paths %r",
                        [
                            record.message_path
                            for record in message_path_repr_map.message_paths
                        ],
                    )

        # Schedule a cleanup of the cache_dir to remove any old assets.
        atexit.register(
            TopicDataService.garbage_collect_old_topic_data, cache_dir=cache_dir
        )

    def __download_roboto_file(
        self,
        representation: RepresentationRecord,
        outfile: pathlib.Path,
    ) -> None:
        file_id = representation.association.association_id
        logger.debug("Getting signed url for file '%s'", file_id)
        signed_url_response = self.__roboto_client.get(f"v1/files/{file_id}/signed-url")
        signed_url = signed_url_response.to_dict(json_path=["data", "url"])

        logger.debug("Downloading file '%s' to %s", file_id, outfile)
        urllib.request.urlretrieve(signed_url, str(outfile))

    def __filter_message_paths(
        self,
        seq: collections.abc.Sequence[MessagePathRecord],
        include_paths: typing.Optional[collections.abc.Sequence[str]],
        exclude_paths: typing.Optional[collections.abc.Sequence[str]],
    ) -> collections.abc.Sequence[MessagePathRecord]:
        if not include_paths and not exclude_paths:
            return seq

        filtered = []
        include_paths_set = set(include_paths or [])
        exclude_paths_set = set(exclude_paths or [])
        for message_path_record in seq:
            message_path_parts = message_path_record.message_path.split(".")
            message_path_parents = set(
                ".".join(message_path_parts[:i])
                for i in range(len(message_path_parts), 0, -1)
            )

            if include_paths and message_path_parents.isdisjoint(include_paths_set):
                continue

            if exclude_paths and not message_path_parents.isdisjoint(exclude_paths_set):
                continue

            filtered.append(message_path_record)

        return filtered

    def __representation_out_file(
        self,
        representation: RepresentationRecord,
        cache_dir: pathlib.Path,
    ) -> pathlib.Path:
        return cache_dir / OUTFILE_NAME_PATTERN.format(
            repr_id=representation.representation_id,
            file_id=representation.association.association_id,
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .actions import (
    ActionParameterArg,
    ActionReferenceParser,
    ActionTimeoutArg,
    add_action_reference_arg,
    add_compute_requirements_args,
    add_container_parameters_args,
    parse_compute_requirements,
    parse_container_overrides,
)
from .orgs import (
    add_org_arg,
    get_defaulted_org_id,
)

__all__ = (
    "ActionParameterArg",
    "ActionReferenceParser",
    "ActionTimeoutArg",
    "add_action_reference_arg",
    "add_compute_requirements_args",
    "add_container_parameters_args",
    "add_org_arg",
    "get_defaulted_org_id",
    "parse_compute_requirements",
    "parse_container_overrides",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import logging
import pathlib
import sys
import typing

from ..config import RobotoConfig
from ..http import (
    BearerTokenDecorator,
    RobotoClient,
    RobotoRequester,
    RobotoTool,
)
from ..version import __version__
from .actions import (
    command_set as actions_command_set,
)
from .argparse import (
    DeprecationAction,
    SortingHelpFormatter,
)
from .collections import (
    command_set as collections_command_set,
)
from .config import check_last_update
from .context import CLIContext
from .datasets import (
    command_set as datasets_command_set,
)
from .devices import (
    command_set as devices_command_set,
)
from .extension import (
    apply_roboto_cli_command_extensions,
    apply_roboto_cli_context_extensions,
)
from .images import (
    command_set as images_command_set,
)
from .invocations import (
    command_set as invocations_command_set,
)
from .orgs import command_set as orgs_command_set
from .tokens import (
    command_set as tokens_command_set,
)
from .triggers import (
    command_set as triggers_command_set,
)
from .users import (
    command_set as users_command_set,
)

COMMAND_SETS = [
    actions_command_set,
    collections_command_set,
    datasets_command_set,
    devices_command_set,
    images_command_set,
    invocations_command_set,
    orgs_command_set,
    users_command_set,
    tokens_command_set,
    triggers_command_set,
]

BETA_USER_POOL_CLIENT_ID = "7p2e45lijin58tuaairtflf3m8"
PROD_USER_POOL_CLIENT_ID = "1gricmdmh0vv582qdd84phab5"


PROGRAMMATIC_ACCESS_BLURB = (
    "To resolve this, please consult the getting started page for programmatic access at "
    + "https://docs.roboto.ai/getting-started/programmatic-access.html."
)


def construct_parser(
    context: typing.Optional[CLIContext] = None,
) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="roboto",
        description=(
            "CLI for interacting with Roboto's Data Platform. "
            "Each of the command groups listed below have their own set of supported subcommands and help pages."
        ),
        formatter_class=SortingHelpFormatter,
    )

    parser.add_argument(
        "--debug",
        help="Deprecated. Please use the --verbose flag instead.",
        deprecation_msg="The --debug flag is deprecated. Please use the --verbose flag instead.",
        action=DeprecationAction,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        help=(
            "Set increasing levels of verbosity. "
            "By default, only ERROR logs are printed. "
            "-v prints WARNING logs, -vv prints INFO logs, -vvv prints DEBUG logs."
        ),
        action="count",
        default=0,
    )

    parser.add_argument(
        "--version",
        help="Show the version of 'roboto' currently running",
        action="store_true",
    )

    parser.add_argument(
        "--profile",
        help="Roboto profile to use; must match a section within the Roboto config.json",
        required=False,
    )

    parser.add_argument(
        "--config-file",
        help="Overrides the location of the roboto config.json file. Defaults to ~/.roboto/config.json",
        type=pathlib.Path,
        required=False,
    )

    parser.add_argument(
        "--suppress-upgrade-check",
        dest="suppress_upgrade_check",
        help="Suppresses the check for a newer 'roboto' package version.",
        action="store_true",
        required=False,
    )

    # https://bugs.python.org/issue29298
    subcommands = parser.add_subparsers(dest="function")

    for command_set in sorted(
        apply_roboto_cli_command_extensions(base_command_sets=COMMAND_SETS),
        key=lambda x: x.name,
    ):
        command_set.sort_commands()
        command_set.add_to_subparsers(subcommands, context)

    return parser


def entry():
    logging.basicConfig(level=logging.ERROR, stream=sys.stderr)
    context = CLIContext()
    parser = construct_parser(context)

    # By default, as soon as we figure out we're running a sub-parser, any field encountered after that subparser
    # will be ignored. This means that `roboto --debug datasets search` will work but
    # `roboto datasets search --debug` will not. In order to work around this, we can use parser_known_args,
    # which gives us back the list of un-evaluated args, and then take a second pass at parse_args with those args.
    # This will only work if our subparsers never re-define top level parameters like --debug, --config-file, etc.
    #
    # This solution was based on a stack overflow post about this issue:
    # https://stackoverflow.com/questions/46962065/add-top-level-argparse-arguments-after-subparser-args
    args, unparsed = parser.parse_known_args()
    args = parser.parse_args(unparsed, args)

    log_level = max(logging.ERROR - (args.verbose * 10), logging.DEBUG)
    logging.getLogger().setLevel(log_level)

    try:
        if args.version:
            print(__version__)
        elif "func" in args:
            __populate_context(
                context=context, parser=parser, profile_override=args.profile
            )
            apply_roboto_cli_context_extensions(base_context=context)

            args.func(args)
        else:
            parser.print_help()
    finally:
        if not args.suppress_upgrade_check:
            check_last_update()


def __populate_context(
    context: CLIContext,
    parser: argparse.ArgumentParser,
    profile_override: typing.Optional[str] = None,
):
    try:
        config = RobotoConfig.from_env(profile_override=profile_override)
    except Exception as exc:
        parser.error(str(exc))

    auth_decorator = BearerTokenDecorator(config.api_key)

    context.roboto_service_base_url = config.endpoint
    context.roboto_client = RobotoClient(
        endpoint=config.endpoint,
        auth_decorator=auth_decorator,
    )
    context.http_client = context.roboto_client.http_client
    context.http_client.set_requester(RobotoRequester.for_tool(RobotoTool.Cli))

    context.extensions = {}



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing
from typing import Optional, Union

from ...auth import (
    EditAccessRequest,
    GetAccessResponse,
)
from ...http import RobotoClient
from ...query import QuerySpecification
from ...sentinels import (
    NotSet,
    NotSetType,
    remove_not_set,
)
from .operations import (
    CreateCollectionRequest,
    UpdateCollectionRequest,
)
from .record import (
    CollectionChangeRecord,
    CollectionContentMode,
    CollectionRecord,
    CollectionResourceRef,
    CollectionResourceType,
)


class Collection:
    """
    A higher-level container for grouping datasets together. Collections can also be used
    to group files from several distinct datasets together.
    """

    __record: CollectionRecord
    __roboto_client: RobotoClient

    @classmethod
    def from_id(
        cls,
        collection_id: str,
        version: Optional[int] = None,
        content_mode: CollectionContentMode = CollectionContentMode.Full,
        roboto_client: typing.Optional["RobotoClient"] = None,
    ) -> "Collection":
        roboto_client = RobotoClient.defaulted(roboto_client)

        query: dict[str, typing.Any] = {"content_mode": content_mode.value}

        if version is not None:
            query["version"] = version

        record = roboto_client.get(
            f"v1/collections/id/{collection_id}", query=query
        ).to_record(CollectionRecord)

        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def create(
        cls,
        description: Optional[str] = None,
        name: Optional[str] = None,
        resources: Optional[list[CollectionResourceRef]] = None,
        dataset_ids: typing.Optional[collections.abc.Collection[str]] = None,
        file_ids: typing.Optional[collections.abc.Collection[str]] = None,
        tags: Optional[list[str]] = None,
        roboto_client: typing.Optional["RobotoClient"] = None,
        caller_org_id: Optional[str] = None,
    ) -> "Collection":
        roboto_client = RobotoClient.defaulted(roboto_client)

        normalized_resources = resources or []
        if dataset_ids:
            normalized_resources.extend(
                [
                    CollectionResourceRef(
                        resource_type=CollectionResourceType.Dataset,
                        resource_id=dataset_id,
                    )
                    for dataset_id in dataset_ids
                ]
            )

        if file_ids:
            normalized_resources.extend(
                [
                    CollectionResourceRef(
                        resource_type=CollectionResourceType.File, resource_id=file_id
                    )
                    for file_id in file_ids
                ]
            )

        request = CreateCollectionRequest(
            name=name,
            description=description,
            resources=normalized_resources,
            tags=tags,
        )

        record = roboto_client.post(
            "v1/collections/create", data=request, caller_org_id=caller_org_id
        ).to_record(CollectionRecord)

        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def list_all(
        cls,
        roboto_client: typing.Optional["RobotoClient"] = None,
        owner_org_id: Optional[str] = None,
        content_mode: CollectionContentMode = CollectionContentMode.SummaryOnly,
    ) -> collections.abc.Generator["Collection", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        spec = QuerySpecification()
        query_params = {"content_mode": content_mode.value}

        while True:
            paginated_result = roboto_client.post(
                "v1/collections/search",
                query=query_params,
                data=spec.model_dump(mode="json"),
                owner_org_id=owner_org_id,
                idempotent=True,
            ).to_paginated_list(CollectionRecord)

            for record in paginated_result.items:
                yield cls(record=record, roboto_client=roboto_client)
            if paginated_result.next_token:
                spec.after = paginated_result.next_token
            else:
                break

    def __init__(
        self,
        record: CollectionRecord,
        roboto_client: typing.Optional["RobotoClient"] = None,
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def collection_id(self) -> str:
        return self.__record.collection_id

    @property
    def datasets(self) -> list[str]:
        return [
            resource.get("resource_id", resource.get("dataset_id"))
            for resource in self.__record.resources.get(
                CollectionResourceType.Dataset, []
            )
        ]

    @property
    def files(self) -> list[str]:
        return [
            resource.get("resource_id", resource.get("file_id"))
            for resource in self.__record.resources.get(CollectionResourceType.File, [])
        ]

    @property
    def record(self) -> CollectionRecord:
        return self.__record

    def add_dataset(self, dataset_id: str) -> "Collection":
        return self.update(
            add_resources=[
                CollectionResourceRef(
                    resource_id=dataset_id, resource_type=CollectionResourceType.Dataset
                )
            ]
        )

    def add_file(self, file_id: str) -> "Collection":
        return self.update(
            add_resources=[
                CollectionResourceRef(
                    resource_id=file_id, resource_type=CollectionResourceType.File
                )
            ]
        )

    def changes(
        self, from_version: Optional[int] = None, to_version: Optional[int] = None
    ) -> collections.abc.Generator["CollectionChangeRecord", None, None]:
        query: dict[str, typing.Any] = {}

        if from_version:
            query["from_version"] = from_version

        if to_version:
            query["to_version"] = to_version

        # Currently this only returns a single page
        paginated_results = self.__roboto_client.get(
            f"v1/collections/id/{self.collection_id}/changes", query=query
        ).to_paginated_list(CollectionChangeRecord)

        for record in paginated_results.items:
            yield record

    def delete(self):
        self.__roboto_client.delete(f"v1/collections/id/{self.collection_id}")

    def get_access(self) -> GetAccessResponse:
        return self.__roboto_client.get(
            f"v1/collections/{self.collection_id}/access"
        ).to_record(GetAccessResponse)

    def edit_access(self, edit: EditAccessRequest) -> GetAccessResponse:
        return self.__roboto_client.put(
            f"v1/collections/{self.collection_id}/access", data=edit
        ).to_record(GetAccessResponse)

    def remove_dataset(self, dataset_id: str) -> "Collection":
        return self.update(
            remove_resources=[
                CollectionResourceRef(
                    resource_id=dataset_id, resource_type=CollectionResourceType.Dataset
                )
            ]
        )

    def remove_file(self, file_id: str) -> "Collection":
        return self.update(
            remove_resources=[
                CollectionResourceRef(
                    resource_id=file_id, resource_type=CollectionResourceType.File
                )
            ]
        )

    def update(
        self,
        add_resources: Union[list[CollectionResourceRef], NotSetType] = NotSet,
        add_tags: Union[list[str], NotSetType] = NotSet,
        description: Optional[Union[NotSetType, str]] = NotSet,
        name: Optional[Union[NotSetType, str]] = NotSet,
        remove_resources: Union[list[CollectionResourceRef], NotSetType] = NotSet,
        remove_tags: Union[list[str], NotSetType] = NotSet,
    ) -> "Collection":
        request = remove_not_set(
            UpdateCollectionRequest(
                name=name,
                description=description,
                add_tags=add_tags,
                remove_tags=remove_tags,
                add_resources=add_resources,
                remove_resources=remove_resources,
            )
        )

        self.__record = self.__roboto_client.put(
            f"v1/collections/id/{self.collection_id}", data=request
        ).to_record(CollectionRecord)

        return self



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum
import os
import re
import typing

import pydantic
import pydantic_settings

ROBOTO_ENV_VAR_PREFIX = "ROBOTO_"


def resolve_env_variables(value: str):
    """
    Given any input string, resolves any environment variables in the string against the current environment.
    """
    resolved_value = value

    # Left side of | is $MY_ENV_VAR style, right side is ${MY_ENV_VAR}
    #
    # Each side has the actual variable name inside an inner capture group (separate from $ or ${}) to make it easier
    # to pass to os.getenv
    pattern = re.compile(r"(\$(\w+))|(\$\{(\w+)})")
    matches = pattern.findall(resolved_value)

    for match in matches:
        left, left_word, right, right_word = match

        if left != "" and left_word != "":
            resolved_value = resolved_value.replace(left, os.getenv(left_word, ""))

        if right != "" and right_word != "":
            resolved_value = resolved_value.replace(right, os.getenv(right_word, ""))

    return resolved_value


class RobotoEnvKey(str, enum.Enum):
    """Environment variable keys available within an action invocation runtime"""

    ActionParametersFile = f"{ROBOTO_ENV_VAR_PREFIX}ACTION_PARAMETERS_FILE"
    ActionTimeout = f"{ROBOTO_ENV_VAR_PREFIX}ACTION_TIMEOUT"
    ApiKey = f"{ROBOTO_ENV_VAR_PREFIX}API_KEY"
    BearerToken = f"{ROBOTO_ENV_VAR_PREFIX}BEARER_TOKEN"
    """Interchangable with ApiKey"""
    ConfigFile = f"{ROBOTO_ENV_VAR_PREFIX}CONFIG_FILE"
    DatasetMetadataChangesetFile = (
        f"{ROBOTO_ENV_VAR_PREFIX}DATASET_METADATA_CHANGESET_FILE"
    )
    DatasetId = f"{ROBOTO_ENV_VAR_PREFIX}DATASET_ID"
    FileMetadataChangesetFile = f"{ROBOTO_ENV_VAR_PREFIX}FILE_METADATA_CHANGESET_FILE"
    InputDir = f"{ROBOTO_ENV_VAR_PREFIX}INPUT_DIR"
    InvocationId = f"{ROBOTO_ENV_VAR_PREFIX}INVOCATION_ID"
    OrgId = f"{ROBOTO_ENV_VAR_PREFIX}ORG_ID"
    OutputDir = f"{ROBOTO_ENV_VAR_PREFIX}OUTPUT_DIR"
    Profile = f"{ROBOTO_ENV_VAR_PREFIX}PROFILE"
    RobotoEnv = f"{ROBOTO_ENV_VAR_PREFIX}ENV"
    RobotoServiceUrl = f"{ROBOTO_ENV_VAR_PREFIX}SERVICE_URL"
    """Deprecated, use RobotoServiceEndpoint instead. Left here until 0.3.3 is released so we can migrate
    existing actions to use the new env var."""
    RobotoServiceEndpoint = f"{ROBOTO_ENV_VAR_PREFIX}SERVICE_ENDPOINT"

    @staticmethod
    def for_parameter(param_name: str) -> str:
        return f"{ROBOTO_ENV_VAR_PREFIX}PARAM_{param_name}"


_roboto_env_instance: typing.Optional["RobotoEnv"] = None


# You'll notice that the alias values here are duplicates of the RobotoEnvKey values. This is not ideal, but is
# necessary for type checking, because the value of alias needs to be a string literal and not a de-referenced variable.
# Even using f-strings will break the type checking.
class RobotoEnv(pydantic_settings.BaseSettings):
    """Enivronment within an action invocation runtime"""

    @classmethod
    def default(cls) -> "RobotoEnv":
        # We need to keep this module level vs. as a class variable, because otherwise pydantic will cast it to a
        # ModelPrivateAttr
        global _roboto_env_instance
        if _roboto_env_instance is None:
            _roboto_env_instance = cls()
        return _roboto_env_instance

    action_parameters_file: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_ACTION_PARAMETERS_FILE"
    )

    action_timeout: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_ACTION_TIMEOUT"
    )

    api_key: typing.Optional[str] = pydantic.Field(
        default=None,
        validation_alias=pydantic.AliasChoices("ROBOTO_API_KEY", "ROBOTO_BEARER_TOKEN"),
    )

    config_file: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_CONFIG_FILE"
    )
    """
    An override for the location of a roboto config file. If not provided, the default ~/.roboto/config.json will be
    used (subject to RobotoConfig's implementation)
    """

    dataset_id: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_DATASET_ID"
    )

    dataset_metadata_changeset_file: typing.Optional[str] = pydantic.Field(
        default=None,
        alias="ROBOTO_DATASET_METADATA_CHANGESET_FILE",
    )

    file_metadata_changeset_file: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_FILE_METADATA_CHANGESET_FILE"
    )

    input_dir: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_INPUT_DIR"
    )

    invocation_id: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_INVOCATION_ID"
    )

    org_id: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_ORG_ID")

    output_dir: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_OUTPUT_DIR"
    )

    profile: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_PROFILE")
    """
    The profile name to use if getting RobotoConfig from a config file.
    """

    roboto_env: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_ENV")

    roboto_service_url: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_SERVICE_URL"
    )
    """Deprecated, use roboto_service_endpoint instead. Left here until 0.3.3 is released so we can migrate
    existing actions to use the new env var."""

    roboto_service_endpoint: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_SERVICE_ENDPOINT"
    )
    """A Roboto Service API endpoint to send requests to, typically https://api.roboto.ai"""



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...image_registry import ImageRegistry
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def list_repos(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    image_registry = ImageRegistry(context.roboto_client)
    paginated_results = image_registry.list_repositories(org_id=args.org)
    while True:
        for repo in paginated_results.items:
            print(json.dumps(repo.model_dump(), indent=2))
        if paginated_results.next_token:
            paginated_results = image_registry.list_repositories(
                page_token=paginated_results.next_token, org_id=args.org
            )
        else:
            break


def list_repos_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)


ls_repos_command = RobotoCommand(
    name="list-repos",
    logic=list_repos,
    setup_parser=list_repos_parser,
    command_kwargs={
        "help": "List image repositories hosted in Roboto's image registry."
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import time
import typing

Condition = typing.Callable[..., bool]
Interval = typing.Union[int, typing.Callable[[int], int]]


class TimeoutError(Exception):
    msg: str

    def __init__(self, msg: str, *args) -> None:
        super().__init__(*args)
        self.msg = msg


def wait_for(
    condition: Condition,
    args: typing.Optional[collections.abc.Sequence[typing.Any]] = None,
    timeout: float = 60 * 5,
    interval: Interval = 5,
    timeout_msg: str = "wait_for timed out",
) -> None:
    """
    Wait for a condition to be truthy.

    Args:
        condition: The condition to wait for. This should be a callable that returns a boolean.
        args: The arguments to pass to the condition callable on each iteration.
        timeout: The maximum amount of time to wait for the condition to be truthy.
        interval: The amount of time to wait between iterations. This can be an integer or a callable.
            If it is a callable, it will be called with the iteration number and should return an integer.
        timeout_msg: The message to include in the TimeoutError if the timeout is reached.

    Raises:
        TimeoutError: If the timeout is reached before the condition is truthy.

    Returns:
        None
    """
    time_remaining = timeout
    args = args if args is not None else []
    iteration = 0
    while time_remaining > 0:
        iteration_start = time.monotonic()
        if iteration > 0:
            if callable(interval):
                time.sleep(interval(iteration))
            else:
                time.sleep(interval)
        if condition(*args):
            return

        iteration_duration = time.monotonic() - iteration_start
        time_remaining -= iteration_duration
        iteration += 1

    raise TimeoutError(timeout_msg)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic


class CreateTokenRequest(pydantic.BaseModel):
    """Request payload to create a new token"""

    expiry_days: int = pydantic.Field(
        description="Number of days until the token expires"
    )
    name: str = pydantic.Field(description="A human-readable name for this token.")
    description: typing.Optional[str] = pydantic.Field(
        default=None, description="An optional longer description for this token."
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import typing

import pydantic

from ..users import UserRecord


class OrgStatus(str, enum.Enum):
    """
    Org status enum
    """

    Provisioning = "provisioning"
    Active = "active"
    Deprovisioning = "de-provisioning"


class OrgTier(str, enum.Enum):
    """
    See our pricing page for details on different organization tiers.
    """

    free = "free"
    premium = "premium"


class OrgRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an organization.
    """

    org_id: str
    name: str
    tier: OrgTier
    status: OrgStatus
    created: datetime.datetime
    created_by: typing.Optional[str] = None


class OrgRoleName(str, enum.Enum):
    """
    Org role enum
    """

    user = "user"
    admin = "admin"
    owner = "owner"


class OrgUserRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an organization user.
    """

    user: UserRecord
    org: OrgRecord
    roles: list[OrgRoleName]


class OrgInviteRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an organization invite.
    """

    invite_id: str
    user_id: str
    invited_by: UserRecord
    org: OrgRecord



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum
import typing

import pydantic
from pydantic import ConfigDict

from .conditions import Condition, ConditionGroup


class SortDirection(str, enum.Enum):
    """The direction to sort the results of a query."""

    Ascending = "ASC"
    Descending = "DESC"

    @staticmethod
    def from_string(value: str) -> "SortDirection":
        lowered = value.lower()

        if lowered == "asc":
            return SortDirection.Ascending
        elif lowered == "desc":
            return SortDirection.Descending
        else:
            raise ValueError(f"Unrecognized sort direction '{value}'")


class QuerySpecification(pydantic.BaseModel):
    """
    Model for specifying a query to the Roboto Platform.

    Examples:
        Specify a query with a single condition:
            >>> from roboto import query
            >>> query_spec = query.QuerySpecification(
            ...     condition=query.Condition(
            ...         field="name",
            ...         comparator=query.Comparator.Equals,
            ...         value="Roboto"
            ...     )
            ... )

        Specify a query with multiple conditions:
            >>> from roboto import query
            >>> query_spec = query.QuerySpecification(
            ...     condition=query.ConditionGroup(
            ...         operator=query.ConditionOperator.And,
            ...         conditions=[
            ...             query.Condition(
            ...                 field="name",
            ...                 comparator=query.Comparator.Equals,
            ...                 value="Roboto"
            ...             ),
            ...             query.Condition(
            ...                 field="age",
            ...                 comparator=query.Comparator.GreaterThan,
            ...                 value=18
            ...             )
            ...         ]
            ...     )
            ... )

        Arbitrarily nest condition groups:
            >>> from roboto import query
            >>> query_spec = query.QuerySpecification(
            ...     condition=query.ConditionGroup(
            ...         operator=query.ConditionOperator.And,
            ...         conditions=[
            ...             query.Condition(
            ...                 field="name",
            ...                 comparator=query.Comparator.Equals,
            ...                 value="Roboto"
            ...             ),
            ...             query.ConditionGroup(
            ...                 operator=query.ConditionOperator.Or,
            ...                 conditions=[
            ...                     query.Condition(
            ...                         field="age",
            ...                         comparator=query.Comparator.GreaterThan,
            ...                         value=18
            ...                     ),
            ...                     query.Condition(
            ...                         field="age",
            ...                         comparator=query.Comparator.LessThan,
            ...                         value=30
            ...                     )
            ...                 ]
            ...             )
            ...         ]
            ...     )
            ... )
    """

    condition: typing.Optional[typing.Union[Condition, ConditionGroup]] = None
    limit: int = 1000
    after: typing.Optional[str] = None  # An encoded PaginationToken
    sort_by: typing.Optional[str] = None
    sort_direction: typing.Optional[SortDirection] = None
    model_config = ConfigDict(extra="forbid")

    def fields(self) -> set[str]:
        """Return a set of all fields referenced in the query."""
        fields = set()

        def _iterconditions(
            condition: typing.Optional[typing.Union[Condition, ConditionGroup]]
        ):
            if condition is None:
                return

            if isinstance(condition, Condition):
                fields.add(condition.field)
            else:
                for cond in condition.conditions:
                    _iterconditions(cond)

        _iterconditions(self.condition)
        return fields



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .cli_extension import (
    RobotoCLIExtension,
    apply_roboto_cli_command_extensions,
    apply_roboto_cli_context_extensions,
)

__all__ = [
    "RobotoCLIExtension",
    "apply_roboto_cli_command_extensions",
    "apply_roboto_cli_context_extensions",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


def safe_access_array(array, index):
    if array is not None and isinstance(array, list) and len(array) > index:
        return array[index]
    return None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import decimal
import typing

# 2038-01-19Z
MAX_32BIT_EPOCH_SECONDS = 2_147_483_647
NSEC_PER_SEC = 1_000_000_000


Time: typing.TypeAlias = typing.Union[int, decimal.Decimal, str, datetime.datetime]


def to_epoch_nanoseconds(value: Time):
    """
    Takes a time value in any of the following formats, and converts it to unix epoch nanoseconds:

    * datetime.datetime
    * int: Unix epoch nanoseconds
    * str: ROS formatted timestamp in the form of "<sec>.<nsec>"
    * decimal.Decimal: ROS formatted timestamp in the form of decimal.Decimal("<sec>.<nsec>")
    """
    if isinstance(value, int):
        if value < 0:
            raise ValueError(
                f"Cannot convert a negative number to epoch nanoseconds, got {value}"
            )
        else:
            return value

    elif isinstance(value, decimal.Decimal):
        # Assume value is a ROS formatted timestamp, <sec>.<nsec>
        return int(value * NSEC_PER_SEC)

    elif isinstance(value, str):
        # Assume value is a ROS formatted timestamp, <sec>.<nsec>
        return int(decimal.Decimal(value) * NSEC_PER_SEC)

    elif isinstance(value, datetime.datetime):
        epoch = datetime.datetime(1970, 1, 1)
        return int((value - epoch).total_seconds() * 1e9)

    else:
        raise TypeError(
            "Input must be either an int (epoch nanoseconds), string (`<sec>.<nsec>`) or a datetime.datetime object"
        )


def utcnow() -> datetime.datetime:
    """Return timezone aware datetime.datetime object, now in UTC."""
    return datetime.datetime.now(tz=datetime.timezone.utc)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain import actions
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def delete(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    action = actions.Action.from_name(
        name=args.action,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    action.delete()
    print(f"Deleted action '{args.action}'")


def delete_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "action",
        metavar="action_reference: <action_name>",
        help="Exact name of action to delete.",
    )
    add_org_arg(parser=parser)


delete_command = RobotoCommand(
    name="delete",
    logic=delete,
    setup_parser=delete_parser,
    command_kwargs={"help": "Delete action and all of its related subresources."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import datetime
import pathlib
import typing

import pydantic

from ..domain.datasets import CreateDatasetRequest

UPLOAD_COMPLETE_FILENAME = ".roboto_upload_complete.json"
UPLOAD_IN_PROGRESS_FILENAME = ".roboto_upload_in_progress.json"


class UploadConfigFileDatasetSection(CreateDatasetRequest):
    """Upload agent config file dataset section"""

    org_id: typing.Optional[str] = None


class UploadConfigFileUploadSection(pydantic.BaseModel):
    """Upload agent config file upload section"""

    delete_uploaded_files: typing.Optional[bool] = None
    exclude_patterns: typing.Optional[typing.List[str]] = None
    include_patterns: typing.Optional[typing.List[str]] = None


class UploadConfigFile(pydantic.BaseModel):
    """Upload agent config file"""

    version: typing.Literal["v1"] = "v1"

    dataset: UploadConfigFileDatasetSection = pydantic.Field(
        default_factory=UploadConfigFileDatasetSection
    )
    upload: UploadConfigFileUploadSection = pydantic.Field(
        default_factory=UploadConfigFileUploadSection
    )


class UploadAgentConfig(pydantic.BaseModel):
    """Upload agent configuration"""

    version: typing.Literal["v1"] = "v1"

    default_org_id: typing.Optional[str] = None
    """
    The org ID to use when creating datasets via --merge-uploads, or other operations which may have an ambiguous org.
    """

    delete_empty_directories: bool = False
    """
    If set to true, directories which are empty (or only contain a .roboto_upload_complete.json) after being uploaded
    will be automatically deleted. This is most useful if combined with delete_upload_files=True
    """

    delete_uploaded_files: bool = False
    """
    If set to true, will delete files from disk after they've been successfully uploaded to Roboto.
    """

    search_paths: list[pathlib.Path]
    """
    Directories to recursively scan for files to upload.
    """

    upload_config_filename: str = ".roboto_upload.json"
    """
    The name of the upload marker file to look for.
    """


class UploadInProgressFile(pydantic.BaseModel):
    """Upload agent in progress file"""

    version: typing.Literal["v1"] = "v1"
    dataset_id: str
    started: datetime.datetime


class UploadCompleteFile(pydantic.BaseModel):
    """Upload agent complete file"""

    version: typing.Literal["v1"] = "v1"
    dataset_id: str
    completed: datetime.datetime



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

# This file is adapted from pandas.compat._optional.py
# https://github.com/pandas-dev/pandas/blob/2419343bfea5dba678146139ca9663d831c47b22/pandas/compat/_optional.py
# As of time of writing (2024-09-20), pandas is distributed under the terms of the BSD 3-Clause License,
# which can be found at:
# https://github.com/pandas-dev/pandas/blob/2419343bfea5dba678146139ca9663d831c47b22/LICENSE

import importlib
import types
import typing
import warnings

IMPORT_ERROR_MSG_TEMPLATE = (
    "Missing optional dependency '{module_name}'. "
    "Re-install roboto using pip or conda with 'roboto[{pip_extra}]' to install a compatible version."
)

KNOWN_PIP_EXTRAS: typing.TypeAlias = typing.Literal["analytics"]


@typing.overload
def import_optional_dependency(
    module_name: str,
    pip_extra: KNOWN_PIP_EXTRAS,
    *,
    errors: typing.Literal["raise"] = ...,
) -> types.ModuleType: ...


@typing.overload
def import_optional_dependency(
    module_name: str,
    pip_extra: KNOWN_PIP_EXTRAS,
    *,
    errors: typing.Literal["warn", "ignore"],
) -> types.ModuleType | None: ...


def import_optional_dependency(
    module_name: str,
    pip_extra: KNOWN_PIP_EXTRAS,
    *,
    errors: typing.Literal["raise", "warn", "ignore"] = "raise",
) -> types.ModuleType | None:
    """
    Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice
    message will be raised.

    Parameters
    ----------
    module_name : str
        The module name.
    pip_extra : str
        The pip extra to use in the ImportError message.
    errors : str {'raise', 'warn', 'ignore'}
        What to do when a dependency is not found or its version is too old.

        * raise : Raise an ImportError
        * ignore: If the module is not installed, return None, otherwise,
          return the module.
    Returns
    -------
    maybe_module : Optional[ModuleType]
        The imported module, when found.
        None is returned when the package is not found and `errors`
        is False, or when the package's version is too old and `errors`
        is ``'warn'`` or ``'ignore'``.
    """
    assert errors in {"raise", "warn", "ignore"}

    msg = IMPORT_ERROR_MSG_TEMPLATE.format(module_name=module_name, pip_extra=pip_extra)
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        if errors == "raise":
            raise ImportError(msg) from None

        if errors == "warn":
            warnings.warn(msg)

        return None

    return module



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import os
import typing

from ...domain import orgs
from ...exceptions import (
    RobotoNoOrgProvidedException,
)

ORG_ARG_HELP = (
    "The calling organization ID. Gets set implicitly if in a single org. "
    + "The `ROBOTO_ORG_ID` environment variable can be set to control the default value."
)

DEFAULT_ORG_ID = os.getenv("ROBOTO_ORG_ID")


def add_org_arg(parser: argparse.ArgumentParser, arg_help: str = ORG_ARG_HELP):
    parser.add_argument(
        "--org", required=False, type=str, help=arg_help, default=DEFAULT_ORG_ID
    )


def get_defaulted_org_id(org_id: typing.Optional[str]) -> str:
    if org_id is not None:
        return org_id

    user_orgs = orgs.Org.for_self()
    if len(user_orgs) == 0:
        raise RobotoNoOrgProvidedException(
            "Current user is not a member of any orgs, and did not provide a --org"
        )
    elif len(user_orgs) > 1:
        raise RobotoNoOrgProvidedException(
            f"Current user is a member of {len(user_orgs)} orgs, and must specify one with --org"
        )
    else:
        return user_orgs[0].org_id



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pydantic


class ContainerUploadCredentials(pydantic.BaseModel):
    """Container upload credentials"""

    username: str
    password: str
    registry_url: str
    image_uri: str


class CreateImageRepositoryRequest(pydantic.BaseModel):
    """Request payload to create an image repository"""

    repository_name: str
    immutable_image_tags: bool


class CreateImageRepositoryResponse(pydantic.BaseModel):
    """Response payload to create an image repository"""

    repository_name: str
    repository_uri: str


class DeleteImageRequest(pydantic.BaseModel):
    """Request payload to delete an image"""

    image_uri: str


class DeleteImageRepositoryRequest(pydantic.BaseModel):
    """Request payload to delete an image repository"""

    repository_name: str
    """The name of the repository to delete."""

    force: bool = False
    """Delete all images in the repository before deleting the repository if the repository is not empty."""


class RepositoryContainsImageResponse(pydantic.BaseModel):
    """Response payload to repository contains image"""

    contains_image: bool


class SetImageRepositoryImmutableTagsRequest(pydantic.BaseModel):
    """Request payload to set image repository tags"""

    repository_name: str
    immutable_image_tags: bool



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import json
import pathlib
import typing

import pydantic

from .env import RobotoEnv
from .logging import default_logger

logger = default_logger()


ROBOTO_API_ENDPOINT = "https://api.roboto.ai"


DEFAULT_ROBOTO_DIR = pathlib.Path.home() / ".roboto"
DEFAULT_ROBOTO_CONFIG_DIR = DEFAULT_ROBOTO_DIR / "config.json"
DEFAULT_ROBOTO_PROFILE_NAME = "default"


_CONFIG_ERROR_SUFFIX = (
    "For more information on setting up the config file used by Roboto's first-party tools, please refer to "
    + "https://docs.roboto.ai/getting-started/programmatic-access.html."
)


class RobotoConfig(pydantic.BaseModel):
    """
    RobotoConfig captures an `api_key` and `endpoint` required to programmatically
    interact with Roboto. Multiple profiles can be configured if desired.
    """

    api_key: str
    endpoint: str = ROBOTO_API_ENDPOINT

    @classmethod
    def from_env(cls, profile_override: typing.Optional[str] = None) -> "RobotoConfig":
        default_env = RobotoEnv.default()

        # If ROBOTO_API_KEY or ROBOTO_BEARER_TOKEN are provided, use that api_key, and either get the endpoint
        # from the environment, or use the default endpoint
        if default_env.api_key:
            endpoint = default_env.roboto_service_endpoint or ROBOTO_API_ENDPOINT

            return RobotoConfig(api_key=default_env.api_key, endpoint=endpoint)

        # Try to read a Roboto config file, either from an env variable specified location, or the default location
        # ~/.roboto/config.json
        if default_env.config_file:
            config_file = pathlib.Path(default_env.config_file)
        else:
            config_file = DEFAULT_ROBOTO_CONFIG_DIR

        if not config_file.is_file():
            raise FileNotFoundError(
                f"No Roboto config file found at specified path '{config_file}'. This may mean that a "
                + "config file was never created, or that your ROBOTO_CONFIG_FILE override environment "
                + "variable is set incorrectly. "
                + _CONFIG_ERROR_SUFFIX
            )

        try:
            config_file_dict = json.loads(config_file.read_text())
        except json.JSONDecodeError:
            raise ValueError(
                f"Roboto config file at path '{config_file}' is not valid JSON. "
                + _CONFIG_ERROR_SUFFIX
            )

        profile_name: typing.Optional[str] = default_env.profile
        profiles: dict[str, RobotoConfig] = {}

        # First try to interpret it as a V1 new-style config file
        try:
            model = RobotoConfigFileV1.model_validate(config_file_dict)
            profiles = model.profiles
            if model.default_profile is not None and profile_name is None:
                profile_name = model.default_profile

        # If that doesn't work, interpret the config file as a list of profiles (V0 format)
        except pydantic.ValidationError:
            for config_profile_name, config_profile in config_file_dict.items():
                if type(config_profile) is dict:
                    try:
                        profiles[config_profile_name] = (
                            RobotoConfigFileProfileV0.model_validate(
                                config_profile
                            ).to_config()
                        )
                    except pydantic.ValidationError:
                        pass

        if len(profiles) == 0:
            raise ValueError(
                f"No user profiles found in config file '{config_file}'. "
                + _CONFIG_ERROR_SUFFIX
            )

        # If a profile name was explicitly passed to this function (for example when called in the CLI's entry.py),
        # that blows over anything else we've seen. Otherwise, use the profile name extracted from either
        # an env variable or the default_profile_name param of a config file.
        #
        # If none of those match, just fall back to the default profile name.
        profile_name = profile_override or profile_name or DEFAULT_ROBOTO_PROFILE_NAME

        if profile_name not in profiles.keys():
            raise ValueError(
                f"User profile '{profile_name}' was not found in config file '{config_file}'. "
                + _CONFIG_ERROR_SUFFIX
            )

        return profiles[profile_name]


class RobotoConfigFileProfileV0(pydantic.BaseModel):
    """V0 Roboto configuration file"""

    token: str
    default_endpoint: str = ROBOTO_API_ENDPOINT

    def to_config(self) -> RobotoConfig:
        return RobotoConfig(api_key=self.token, endpoint=self.default_endpoint)


class RobotoConfigFileV1(pydantic.BaseModel):
    """V1 Roboto configuration file"""

    version: typing.Literal["v1"]
    profiles: dict[str, RobotoConfig]
    default_profile: typing.Optional[str] = DEFAULT_ROBOTO_PROFILE_NAME

    @pydantic.model_validator(mode="after")
    def validate(self):
        if len(self.profiles) == 0:
            raise ValueError("No profiles found in config file.")

        if (
            self.default_profile or DEFAULT_ROBOTO_PROFILE_NAME
        ) not in self.profiles.keys():
            raise ValueError(
                f"Default profile '{self.default_profile}' was not found in config file."
            )

        return self



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import enum
import hashlib
import json
import typing

import pydantic

from roboto.pydantic.serializers import (
    field_serializer_user_metadata,
)
from roboto.types import UserMetadata


class Accessibility(str, enum.Enum):
    """
    Controls who can query for and invoke an action.

    Future accessibility levels may include: "user" and/or "team".
    """

    Organization = "organization"
    """All members of the organization owning the Action can query for and invoke the action."""

    ActionHub = "action_hub"
    """All users of Roboto can query for and invoke the action."""


class ActionParameter(pydantic.BaseModel):
    """
    A parameter that can be provided to an Action at invocation time.
    """

    name: str
    required: bool = False
    description: typing.Optional[str] = None
    default: typing.Optional[typing.Any] = None
    """
    Default value applied for parameter if it is not required and no value is given at invocation.
    Accepts any default value, but coerced to a string.
    """

    model_config = pydantic.ConfigDict(extra="forbid")

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, ActionParameter):
            return NotImplemented

        return (
            self.name == other.name
            and self.required == other.required
            and self.description == other.description
            and self.default == other.default
        )

    @pydantic.field_validator("default", mode="before")
    @classmethod
    def validate_default(cls, v: typing.Optional[typing.Any]) -> typing.Optional[str]:
        if v is None:
            return v

        return str(v)


class ActionParameterChangeset(pydantic.BaseModel):
    """
    A changeset used to modify Action parameters.
    """

    put_parameters: list[ActionParameter] = pydantic.Field(default_factory=list)
    remove_parameters: list[str] = pydantic.Field(default_factory=list)

    class Builder:
        __put_parameters: list[ActionParameter]
        __remove_parameters: list[str]

        def __init__(self) -> None:
            self.__put_parameters = []
            self.__remove_parameters = []

        def put_parameter(
            self, parameter: ActionParameter
        ) -> "ActionParameterChangeset.Builder":
            self.__put_parameters.append(parameter)
            return self

        def remove_parameter(
            self, parameter_name: str
        ) -> "ActionParameterChangeset.Builder":
            self.__remove_parameters.append(parameter_name)
            return self

        def build(self) -> "ActionParameterChangeset":
            changeset: collections.abc.Mapping = {
                "put_parameters": self.__put_parameters,
                "remove_parameters": self.__remove_parameters,
            }
            return ActionParameterChangeset(**{k: v for k, v in changeset.items() if v})

    def is_empty(self) -> bool:
        return not self.put_parameters and not self.remove_parameters


class ActionReference(pydantic.BaseModel):
    """Qualified action reference."""

    name: str
    digest: typing.Optional[str] = None
    owner: typing.Optional[str] = None

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, ActionReference):
            return NotImplemented

        return (
            self.name == other.name
            and self.digest == other.digest
            and self.owner == other.owner
        )

    def __str__(self) -> str:
        return (
            f"{self.owner}/{self.name}"
            if not self.digest
            else f"{self.owner}/{self.name}@{self.digest}"
        )


class ExecutorContainer(enum.Enum):
    """
    Type of container running as part of an action invocation
    """

    LogRouter = "firelens_log_router"
    Monitor = "monitor"
    Setup = "setup"
    Action = "action"
    OutputHandler = "output_handler"


class ComputeRequirements(pydantic.BaseModel):
    """
    Compute requirements for an action invocation.

    .. _Relevant AWS Fargate documentation:
        https://docs.aws.amazon.com/AmazonECS/latest/developerguide/AWS_Fargate.html#fargate-tasks-size
    """

    vCPU: int = 512  # 256, 512, 1024, 2048, 4096, 8192, 16384
    memory: int = 1024  # 512, 1024, 2024, ... (120 * 1024) in MiB
    gpu: typing.Literal[False] = False  # Not yet supported
    storage: int = 21  # in GiB (min 21, max 200 if on premium tier)

    @pydantic.model_validator(mode="after")
    def validate_storage_limit(self):
        if self.storage < 21:
            raise ValueError(
                f"Unsupported Storage value {self.storage}. Storage must be at least 21 GiB."
            )
        return self

    @pydantic.model_validator(mode="after")
    def validate_vcpu_mem_combination(self) -> "ComputeRequirements":
        allowed_vpcu = (256, 512, 1024, 2048, 4096, 8192, 16384)
        if self.vCPU not in allowed_vpcu:
            raise ValueError(f"Unsupported vCPU value. Allowed options: {allowed_vpcu}")

        memory = self.memory
        allowed_memory: collections.abc.Sequence[int] = list()
        if self.vCPU == 256:
            allowed_memory = [512, 1024, 2048]
        elif self.vCPU == 512:
            allowed_memory = range(1024, 5 * 1024, 1024)
        elif self.vCPU == 1024:
            allowed_memory = range(2 * 1024, 9 * 1024, 1024)
        elif self.vCPU == 2048:
            allowed_memory = range(4 * 1024, 17 * 1024, 1024)
        elif self.vCPU == 4096:
            allowed_memory = range(8 * 1024, 31 * 1024, 1024)
        elif self.vCPU == 8192:
            allowed_memory = range(16 * 1024, 61 * 1024, 4 * 1024)
        elif self.vCPU == 16384:
            allowed_memory = range(32 * 1024, 121 * 1024, 8 * 1024)
        else:
            raise ValueError(f"Unknown vCPU value {self.vCPU}")

        if memory not in allowed_memory:
            raise ValueError(
                f"Unsupported memory/vCPU combination, allowed memory for {self.vCPU} vCPU: {list(allowed_memory)}"
            )

        return self

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, ComputeRequirements):
            return NotImplemented

        return (
            self.vCPU == other.vCPU
            and self.memory == other.memory
            and self.gpu == other.gpu
            and self.storage == other.storage
        )

    model_config = pydantic.ConfigDict(extra="forbid")


class ContainerParameters(pydantic.BaseModel):
    """
    Container parameters for an action invocation.
    """

    command: typing.Optional[list[str]] = None
    entry_point: typing.Optional[list[str]] = None
    env_vars: typing.Optional[dict[str, str]] = None
    workdir: typing.Optional[str] = None

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, ContainerParameters):
            return NotImplemented

        return (
            self.command == other.command
            and self.entry_point == other.entry_point
            and self.env_vars == other.env_vars
            and self.workdir == other.workdir
        )

    model_config = pydantic.ConfigDict(extra="forbid")


class ActionRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of an action.
    """

    # Required fields without defaults
    created: datetime.datetime  # Persisted as ISO 8601 string in UTC
    created_by: str
    modified: datetime.datetime  # Persisted as ISO 8601 string in UTC
    modified_by: str
    name: str  # Sort key
    org_id: str  # Partition key

    # Optional fields with defaults
    accessibility: Accessibility = Accessibility.Organization
    compute_requirements: typing.Optional[ComputeRequirements] = None
    container_parameters: typing.Optional[ContainerParameters] = None
    description: typing.Optional[str] = None
    digest: typing.Optional[str] = None
    inherits: typing.Optional[ActionReference] = None
    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    parameters: list[ActionParameter] = pydantic.Field(default_factory=list)
    # Persisted as ISO 8601 string in UTC
    published: typing.Optional[datetime.datetime] = None
    tags: list[str] = pydantic.Field(default_factory=list)
    uri: typing.Optional[str] = None
    short_description: typing.Optional[str] = None
    timeout: typing.Optional[int] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.digest = self.compute_digest() if not self.digest else self.digest

    def compute_digest(self) -> str:
        hasher = hashlib.blake2b(
            digest_size=16,
            # https://docs.python.org/3.9/library/hashlib.html#personalization
            person=b"ActionRecord",
        )
        digestable = self.model_dump(exclude_unset=True, mode="json")
        hasher.update(json.dumps(digestable, sort_keys=True).encode("utf-8"))
        return hasher.hexdigest()

    @property
    def reference(self) -> ActionReference:
        return ActionReference(
            name=self.name,
            digest=self.digest,
            owner=self.org_id,
        )

    @pydantic.field_serializer("metadata")
    def serialize_metadata(self, metadata: dict[str, typing.Any]) -> UserMetadata:
        return field_serializer_user_metadata(metadata)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .http_client import NotificationsClient
from .http_resources import (
    UpdateNotificationRequest,
)
from .record import NotificationRecord
from .validator import (
    EmailLifecycleStatus,
    NotificationChannel,
    NotificationType,
    ReadStatus,
    WebUiLifecycleStatus,
)

__all__ = [
    "NotificationType",
    "NotificationChannel",
    "NotificationsClient",
    "NotificationRecord",
    "ReadStatus",
    "UpdateNotificationRequest",
    "EmailLifecycleStatus",
    "WebUiLifecycleStatus",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import typing

from ...compat import import_optional_dependency
from .match import Match

if typing.TYPE_CHECKING:
    import numpy.typing  # pants: no-infer-dep


class QuerySignal:
    """
    A sequence of values to match against other signals, looking for similar subsequences.

    Query signals must be at minimum 3 values long for results to be meaningful.
    """

    MIN_QUERY_LENGTH: typing.ClassVar = 3

    __query: numpy.typing.NDArray

    @staticmethod
    def from_array(sequence: numpy.typing.NDArray) -> "QuerySignal":
        if len(sequence) < QuerySignal.MIN_QUERY_LENGTH:
            raise ValueError("Query signal must be at least 3 values long")

        return QuerySignal(sequence)

    def __init__(self, query: numpy.typing.NDArray):
        self.__query = query

    def match(
        self,
        target: numpy.typing.NDArray,
        *,
        max_distance: typing.Optional[float] = None,
        max_matches: typing.Optional[int] = None,
        normalize: bool = False,
    ) -> list[Match]:
        """
        Find at least one subsequence of the target signal that is similar to the query signal.
        Even if there is no true similarity between the query and target,
        this will always return at least one :class:`~roboto.query.signal_similarity.Match`.
        Matches are expected to improve in quality as the target is more relevant to the query.
        Matches are returned sorted in ascending order by their distance, with the best matches (lowest distance) first.

        If ``max_distance`` is provided, only matches with a distance less than ``max_distance`` will be returned.
        Defaults to--given distances computed against all comparison windows in the target signal--the maximum of:
            1. the minimum distance
            2. the mean distance minus two standard deviations

        If ``max_matches`` is provided, only the top ``max_matches`` matches will be returned.

        If ``normalize`` is True, values will be projected to the unit scale before matching.
        This is useful if you want to match windows of the target signal regardless of scale.
        For example, a query signal of ``[1., 2., 3.]`` will perfectly match (distance == 0)
        the target ``[1000., 2000., 3000.]`` if ``normalize`` is True,
        but would have a distance of nearly 3800 if ``normalize`` is False.
        """
        if len(self.__query) > len(target):
            raise ValueError("Query signal must be shorter than target signal")

        matches: list[Match] = []
        stumpy = import_optional_dependency("stumpy", "analytics")
        for distance, start_idx in stumpy.match(
            self.__query,
            target,
            max_distance=max_distance,
            max_matches=max_matches,
            normalize=normalize,
        ):
            end_idx = start_idx + len(self.__query)
            matches.append(
                Match(
                    start_idx=start_idx,
                    end_idx=end_idx,
                    distance=distance,
                    subsequence=target[start_idx:end_idx],
                )
            )

        return matches

    def match_all(
        self,
        targets: collections.abc.Iterable[numpy.typing.NDArray],
        *,
        max_distance: typing.Optional[float] = None,
        max_matches: typing.Optional[int] = None,
        normalize: bool = False,
    ) -> list[Match]:
        """
        Find at least one subsequence within the collection of targets that is similar to the query signal.
        See :meth:`~roboto.query.signal_similarity.QuerySignal.match` for more details.
        """
        matches: list[Match] = []
        for target in targets:
            matches.extend(
                self.match(
                    target,
                    max_distance=max_distance,
                    max_matches=max_matches,
                    normalize=normalize,
                )
            )

        matches.sort(key=lambda match: match.distance)

        if max_matches is not None:
            return matches[:max_matches]

        return matches



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing
import urllib.parse

from ...http import RobotoClient
from ...regionalization import RobotoRegion
from .org_invite import OrgInvite
from .org_operations import (
    CreateOrgRequest,
    UpdateOrgRequest,
    UpdateOrgUserRequest,
)
from .org_records import (
    OrgRecord,
    OrgRoleName,
    OrgStatus,
    OrgTier,
    OrgUserRecord,
)


class Org:
    """
    A shared account between many Roboto users. Orgs may have different limits
    and functionality depending on tier.
    """

    __record: OrgRecord
    __roboto_client: RobotoClient

    @classmethod
    def create(
        cls,
        name: str,
        bind_email_domain: bool = False,
        data_region: RobotoRegion = RobotoRegion.US_WEST,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Org":
        roboto_client = RobotoClient.defaulted(roboto_client)
        request = CreateOrgRequest(
            name=name, bind_email_domain=bind_email_domain, data_region=data_region
        )
        record = roboto_client.post("v1/orgs", data=request).to_record(OrgRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def from_id(
        cls, org_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> "Org":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(f"v1/orgs/id/{org_id}").to_record(OrgRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def for_self(
        cls, roboto_client: typing.Optional[RobotoClient] = None
    ) -> collections.abc.Sequence["Org"]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        records = roboto_client.get("v1/orgs/caller").to_record_list(OrgRecord)
        return [cls(record=record, roboto_client=roboto_client) for record in records]

    @classmethod
    def for_user(
        cls, user_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> collections.abc.Sequence["Org"]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        records = roboto_client.get(
            f"v1/users/id/{urllib.parse.quote_plus(user_id)}/orgs"
        ).to_record_list(OrgRecord)
        return [cls(record=record, roboto_client=roboto_client) for record in records]

    def __init__(
        self, record: OrgRecord, roboto_client: typing.Optional[RobotoClient] = None
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def name(self) -> str:
        return self.__record.name

    @property
    def org_id(self):
        return self.__record.org_id

    @property
    def status(self) -> OrgStatus:
        return self.__record.status

    @property
    def tier(self) -> OrgTier:
        return self.__record.tier

    def add_role_for_user(self, user_id: str, role: OrgRoleName) -> "Org":
        self.__roboto_client.put(
            f"v1/orgs/id/{self.org_id}/users/id/{urllib.parse.quote_plus(user_id)}",
            data=UpdateOrgUserRequest(add_roles=[role]),
        )
        return self

    def remove_role_from_user(self, user_id: str, role: OrgRoleName) -> "Org":
        self.__roboto_client.put(
            f"v1/orgs/id/{self.org_id}/users/id/{urllib.parse.quote_plus(user_id)}",
            data=UpdateOrgUserRequest(remove_roles=[role]),
        )
        return self

    def bind_email_domain(self, email_domain: str) -> "Org":
        self.__roboto_client.put(
            f"v1/orgs/id/{self.org_id}/email_domains/id/{urllib.parse.quote_plus(email_domain)}"
        )
        return self

    def delete(self) -> None:
        self.__roboto_client.delete(f"v1/orgs/id/{self.org_id}")

    def email_domains(self) -> collections.abc.Collection[str]:
        return self.__roboto_client.get(
            f"v1/orgs/id/{self.org_id}/email_domains"
        ).to_dict(json_path=["data"])

    def invite_user(
        self,
        user_id: str,
    ) -> OrgInvite:
        return OrgInvite.create(
            invited_user_id=user_id,
            org_id=self.org_id,
            roboto_client=self.__roboto_client,
        )

    def invites(self) -> collections.abc.Collection[OrgInvite]:
        return OrgInvite.for_org(org_id=self.org_id, roboto_client=self.__roboto_client)

    def refresh(self) -> "Org":
        self.__record = self.__roboto_client.get(f"v1/orgs/id/{self.org_id}").to_record(
            OrgRecord
        )
        return self

    def remove_user(self, user_id: str) -> "Org":
        self.__roboto_client.delete(
            f"v1/orgs/id/{self.org_id}/users/id/{urllib.parse.quote_plus(user_id)}"
        )
        return self

    def update(self, update: UpdateOrgRequest) -> "Org":
        self.__record = self.__roboto_client.put(
            f"v1/orgs/id/{self.org_id}", data=update
        ).to_record(OrgRecord)
        return self

    def users(self) -> collections.abc.Collection[OrgUserRecord]:
        return self.__roboto_client.get(
            f"v1/orgs/id/{self.org_id}/users"
        ).to_record_list(OrgUserRecord)

    def unbind_email_domain(self, email_domain: str) -> "Org":
        self.__roboto_client.delete(
            f"v1/orgs/id/{self.org_id}/email_domains/id/{urllib.parse.quote_plus(email_domain)}"
        )
        return self

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")
